# Copyright (C) 2010-2022, kachelaqa <kachelaqa@gmail.com>

import os
from datetime import datetime
from email.utils import parsedate
from pkgbrowser.enum import State, Validation, Backup


class Format(object):
    @staticmethod
    def size(size, precision=1):
        if size < 0:
            return ''
        if size < 1024:
            return '%.0f B' % size
        size /= 1024.0
        if size < 1024:
            return '%.*f KiB' % (precision, size)
        size /= 1024
        if size < 1024 or not precision:
            return '%.*f MiB' % (precision, size)
        return '%.*f GiB' % (precision, size / 1024)

    @staticmethod
    def number(number, precision=2, signed=False):
        if signed or number >= 0:
            if isinstance(number, float):
                return '%.*f' % (precision, number)
            return str(number)
        return ''

    @staticmethod
    def date(date, locale=True):
        result = ''
        if isinstance(date, str):
            date = parsedate(date)
        if date:
            try:
                if isinstance(date, (int, float)):
                    date = datetime.fromtimestamp(date)
                else:
                    date = datetime(*date[:6])
            except (ValueError, TypeError):
                pass
            else:
                if date.year >= 2000:
                    if locale:
                        result = date.strftime('%c')
                    else:
                        result = date.strftime('%Y/%m/%d - %H:%M')
        return result

    def _escape(self, string, normalize=False):
        if isinstance(string, bytes):
            string = string.decode('utf-8')
        elif string is None:
            string = ''
        string = string.replace('&', """&amp;"""
                      ).replace('>', """&gt;"""
                      ).replace('<', """&lt;"""
                      ).replace('"', """&quot;"""
                      )
        if normalize:
            string = """&nbsp;""".join(string.split())
        return string

    def symbols(self):
        return {'angle': '+', 'tee': '+', 'line': '-', 'stem': '|'}

    def colours(self):
        return {
            'body-text': 'black',
            'body-back': 'white',
            'table-border': 'grey',
            'label-text': 'black',
            'label-back': 'white',
            'tree-lines': 'grey',
            }

    def information(self, package, page):
        markup = ["""
            <html><head><style type="text/css">
            body {color: %(body-text)s; background-color: %(body-back)s;}
            table.main {background-color: %(table-border)s;}
            p {margin-top: 25; margin-left: 35; margin-right: 10;}
            p.tree {font-family: monospace; color: %(tree-lines)s;}
            p.tree span {font-family: sans serif; color: %(body-text)s;}
            td.key, p.files, p.tree, p.log {white-space: pre;}
            td.key {padding: 4; padding-right: 15; color: %(label-text)s;
                    background-color: %(label-back)s;}
            td.value {white-space: pre-wrap; padding: 4;
                      background-color: %(body-back)s;}
            table.cache td, table.backup td {padding-right: 20;}
            img {float: left;}
            </style></head><body>
            <table class="main" cellspacing="1">
            """ % self.colours()]
        row = """
            <tr><td class="key">%s</td>
            <td class="value" width="100%%">%s</td></tr>
            """
        link = """<a name="%s" href="%s">%s</a>%s"""
        img = """<img src="%s"/>&nbsp;%s"""
        span = """<span>%s</span>"""
        state = package['state']
        text = None
        rows = []
        rows.append(('name', self.tr('Name')))
        if not state & State.Unknown:
            rows.append(('version', self.tr('Version')))
        if page == 'details' and not state & State.Unknown:
            rows.append(('description', self.tr('Description')))
            rows.append(('url', self.tr('Url')))
            if state & State.AUR:
                rows.append(('aururl', self.tr('AUR Url')))
            else:
                rows.append(('pkgurl', self.tr('Package Url')))
            rows.append(('license', self.tr('Licenses')))
            rows.append(('state', self.tr('Status')))
            rows.append(('repository', self.tr('Repository')))
            if state & State.AUR:
                rows.append(('votes', self.tr('Votes')))
                rows.append(('popularity', self.tr('Popularity')))
                rows.append(('outdated', self.tr('Out Of Date')))
            rows.append(('groups', self.tr('Groups')))
            rows.append(('provides', self.tr('Provides')))
            rows.append(('depends', self.tr('Dependencies')))
            if state & State.AUR:
                rows.append(('makedepends', self.tr('Build Requires')))
            rows.append(('optdepends', self.tr('Optional')))
            rows.append(('required', self.tr('Required By')))
            if not state & State.AUR or state & State.Installed:
                rows.append(('optional', self.tr('Optional For')))
            rows.append(('conflicts', self.tr('Conflicts With')))
            rows.append(('replaces', self.tr('Replaces')))
            rows.append(('arch', self.tr('Architecture')))
            if state & State.AUR:
                rows.append(('maintainer', self.tr('Maintainer')))
                rows.append(('submitted', self.tr('First Submitted')))
                rows.append(('modified', self.tr('Last Updated')))
            else:
                rows.append(('packager', self.tr('Maintainer')))
                rows.append(('built', self.tr('Build Date')))
            if state & State.Installed:
                rows.append(('installed', self.tr('Install Date')))
            elif not state & State.AUR:
                rows.append(('download', self.tr('Download Size')))
            if not state & State.AUR or state & State.Installed:
                rows.append(('size', self.tr('Installed Size')))
            if state & State.Installed:
                rows.append(('script', self.tr('Install Script')))
                rows.append(('validation', self.tr('Validation')))
        else:
            if page == 'tree':
                symbols = self.symbols()
                branch = '%%s%(angle)s%(line)s%(line)s%%s' % symbols
                junction = '%%s%(tee)s%(line)s%(line)s%%s' % symbols
                stem = '%(stem)s ' % symbols
                def process(parent, padding=''):
                    output = []
                    if padding:
                        padding += ' '
                    length = len(parent)
                    count = 0
                    for name, provides, state, depends in parent:
                        count += 1
                        if padding or count > 1:
                            output.append(padding + stem)
                        label = link % (page, name, name, '')
                        if provides is not None:
                            label = '%s (%s)' % (provides, label)
                        if not state & State.Installed:
                            label += '&nbsp;*'
                            if state & State.AUR:
                                label += '*'
                        if count == length:
                            output.append(junction % (padding, span % label))
                            output.extend(process(depends, padding + '  '))
                        else:
                            output.append(branch % (padding, span % label))
                            output.extend(process(depends, padding + stem))
                    return output
                tree = package[page]
                if tree is not None:
                    text = '\n'.join(process(tree['packages']))
                    if text:
                        rows.append(('tree:installed', self.tr('Installed')))
                        rows.append(('tree:missing', self.tr('Missing')))
            elif page == 'cache':
                sources = []
                for items in package[page] or ():
                    for item in items:
                        item = os.path.split(self._escape(item))
                        sources.append(
                            """<tr><td>%s</td><td><i>%s</i></td></tr>""" %
                            (item[1], item[0]))
                if sources:
                    text = """<table class="%s">%s</table>""" % (
                            page, ''.join(sources))
            elif page == 'backup':
                backups = package[page]
                if backups is not None:
                    lines = []
                    for state, path in backups:
                        lines.append(
                            """<tr><td><i>%s</i></td><td>%s</td></tr>""" %
                            (self.backup(state), self._escape(path)))
                    if lines:
                        text = """<table class="%s">%s</table>""" % (
                                page, ''.join(lines))
                    else:
                        text = ''
            else:
                text = package[page]
                if text is not None:
                    text = self._escape(text)
            if not text:
                if text is not None:
                    if page == 'tree':
                        text = self.tr('no dependencies')
                    elif page == 'files':
                        text = self.tr('no files')
                    elif page == 'log':
                        text = self.tr('no log entries')
                    elif page == 'backup':
                        text = self.tr('no backup files')
                if not text:
                    text = self.tr('no data')
                text = """<span><i>-- %s --</i></span>""" % text
            text = """<p class="%s">%s</p><div/>""" % (page, text)
        dates = set((
            'installed', 'built', 'submitted', 'modified', 'outdated',
            ))
        numbers = set((
            'votes', 'popularity',
            ))
        flagged = dates.union((
            'name', 'version', 'description', 'packager',
            ))
        permanent = set((
            'depends', 'required', 'license', 'maintainer',
            ))
        urls = set((
            'url', 'pkgurl', 'aururl',
            ))
        for key, title in rows:
            title = self._escape(title, True)
            if key == 'tree:installed':
                if tree['installed']:
                    value = '%s (%d)' % (
                        self.size(tree['isize'], 0), tree['installed'])
                else:
                    value = self.tr('None')
            elif key == 'tree:missing':
                if tree['missing'] or tree['aur']:
                    value = '%s (%d)' % (
                        self.size(tree['msize'], 0), tree['missing'])
                    if tree['aur']:
                        value += ' + (%d AUR)' % tree['aur']
                else:
                    value = self.tr('None')
            else:
                value = package[key]
                if value and isinstance(value, list):
                    items = []
                    for name, data in value:
                        name = self._escape(name)
                        data = self._escape(data, True)
                        items.append(link % (key, name, name, data))
                    value = ' '.join(items)
                elif key == 'state':
                    value = self._escape(self.status(state))
                    icon = self.icon(state)
                    if icon:
                        value = img % (icon, value)
                elif key == 'validation':
                    value = self._escape(self.validation(value))
                else:
                    if not value and key == 'outdated':
                        value = False
                    if isinstance(value, bool):
                        if value:
                            value = self.tr('Yes')
                        else:
                            value = self.tr('No')
                    else:
                        if key == 'size' or key == 'download':
                            value = self.size(value)
                        elif key in numbers:
                            value = self.number(value)
                        elif key in dates:
                            value = self.date(value)
                        if not value:
                            if key in permanent:
                                value = self.tr('None')
                            elif key in flagged:
                                value = self.tr('Unknown')
                            else:
                                continue
                    value = self._escape(value)
                    if key == 'name':
                        value = """<b>%s</b>""" % value
                    elif key in urls:
                        value = link % ('', value, value, '')
            markup.append(row % (title, value))
        markup.append("""</table>%s</body></html>""" % (text or ''))
        return ''.join(markup)

    def statistics(self, data):
        markup = ["""
            <html><head><style type="text/css">
            table {margin-top: 10; margin-bottom: 10; margin-right: 20}
            </style></head><body>
            <div><table width="320">
            """]
        row = """
            <tr><td><b>%s</b></td>
            <td align="right" width="40%%">%s</td>
            <td align="right">%s</td></tr>
            """
        line = """<tr><td colspan="3"><hr></td></tr>"""
        markup.append(row % (
            """<b>%s</b>""" % self._escape(self.tr('Repository')),
            """<b>%s</b>""" % self._escape(self.tr('Count')),
            """<b>%s</b>""" % self._escape(self.tr('Installed Size')),
            ))
        markup.append(line)
        for key, count, size in data:
            if key == 'total':
                key = self._escape(self.tr('Total'))
                markup.append(line)
            markup.append(row % (key, count, self.size(size, 0)))
        markup.append("""</table></div></body></html>""")
        return ''.join(markup)

    def file(self, text, source):
        text = text.strip()
        if source == 'details':
            from string import whitespace
            spaces = whitespace + '\xa0\ufffc'
            lines = text.splitlines()
            for index in range(0, len(lines), 2):
                lines[index] = '%s: ' % lines[index].strip(spaces)
                lines[index + 1] = '%s\n' % lines[index + 1].strip(spaces)
            text = ''.join(lines)
        else:
            header, data = text.partition('\n\n')[::2]
            header = '-'.join(header.splitlines()[1:4:2])
            data = data.strip().splitlines()
            if source == 'cache' or source == 'backup':
                data = self._columns(list(zip(data[::2], data[1::2])))
            text = '%s\n\n  %s\n' % (header, '\n  '.join(data))
        return text

    def _columns(self, lines, spacing=2, indent=0):
        result = []
        indent = ' ' * indent
        widths = [max(len(value) for value in column) + spacing
                  for column in zip(*lines)]
        for line in lines:
            result.append(indent + ''.join(
                '%-*s' % item for item in zip(widths, line)))
        return result

    def list(self, summaries):
        lines = []
        for summary in summaries:
            info = {
                'n': summary.name,
                'v': summary.version,
                'r': summary.repository,
                }
            if summary.state & State.Group:
                format = '%(r)s/%(n)s\n'
            else:
                format = '%(r)s/%(n)s-%(v)s\n'
            lines.append(format % info)
        return ''.join(lines)

    def status(self, state):
        if state & State.Orphan:
            return self.tr('orphan')
        if state & State.Dependency:
            return self.tr('dependency')
        if state & State.Optional:
            return self.tr('optional')
        if state & (State.Explicit | State.Installed):
            return self.tr('installed')
        if state & State.Upgrade:
            return self.tr('upgrade')
        if state & State.Downgrade:
            return self.tr('downgrade')
        return self.tr('non-installed')

    def backup(self, state):
        if state == Backup.Unmodified:
            return self.tr('unmodified')
        if state == Backup.Modified:
            return self.tr('modified')
        if state == Backup.Missing:
            return self.tr('missing')
        if state == Backup.Unreadable:
            return self.tr('unreadable')
        return self.tr('unknown')

    def validation(self, validation):
        string = ''
        if validation:
            if validation & Validation.Nothing:
                string = self.tr('None')
            else:
                if validation & Validation.Signature:
                    string += self.tr('Signature') + ', '
                if validation & Validation.Md5sum:
                    string += self.tr('MD5 Sum') + ', '
                if validation & Validation.Sha256sum:
                    string += self.tr('SHA256 Sum') + ', '
                string = string[:-2]
        if not string:
            string = self.tr('Unknown')
        return string



# Copyright (C) 2010-2022, kachelaqa <kachelaqa@gmail.com>


class State(object):
    from pkgbrowser import alpm
    NonInstalled = alpm.PKG_STATUS_NONINSTALLED
    Installed = alpm.PKG_STATUS_INSTALLED
    Explicit = alpm.PKG_STATUS_EXPLICIT
    Dependency = alpm.PKG_STATUS_DEPENDENCY
    Optional = alpm.PKG_STATUS_OPTIONAL
    Orphan = alpm.PKG_STATUS_ORPHAN
    Foreign = alpm.PKG_STATUS_FOREIGN
    Upgrade = alpm.PKG_STATUS_UPGRADE
    Downgrade = alpm.PKG_STATUS_DOWNGRADE
    Update = Upgrade | Downgrade
    AUR = alpm.PKG_STATUS_MAX
    Group = alpm.PKG_STATUS_MAX << 1
    Unknown = alpm.PKG_STATUS_MAX << 2
    Database = alpm.PKG_STATUS_MAX << 3
    del alpm

class Validation(object):
    from pkgbrowser import alpm
    Unknown = alpm.PKG_VALIDATION_UNKNOWN
    Nothing = alpm.PKG_VALIDATION_NONE
    Md5sum = alpm.PKG_VALIDATION_MD5SUM
    Sha256sum = alpm.PKG_VALIDATION_SHA256SUM
    Signature = alpm.PKG_VALIDATION_SIGNATURE
    del alpm

class Backup(object):
    from pkgbrowser import alpm
    Unknown = alpm.PKG_BACKUP_UNKNOWN
    Unmodified = alpm.PKG_BACKUP_UNMODIFIED
    Modified = alpm.PKG_BACKUP_MODIFIED
    Missing = alpm.PKG_BACKUP_MISSING
    Unreadable = alpm.PKG_BACKUP_UNREADABLE
    del alpm

class Source(object):
    Local = 1
    Sync = 2
    Foreign = 4
    Group = 8
    AUR = 16



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'designer/about.ui'
#
# Created by: PyQt5 UI code generator 5.15.6
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_AboutDialog(object):
    def setupUi(self, AboutDialog):
        AboutDialog.setObjectName("AboutDialog")
        AboutDialog.resize(451, 334)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/icons/app.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        AboutDialog.setWindowIcon(icon)
        self._central_layout = QtWidgets.QGridLayout(AboutDialog)
        self._central_layout.setObjectName("_central_layout")
        self.icon = QtWidgets.QLabel(AboutDialog)
        self.icon.setPixmap(QtGui.QPixmap(":/icons/app.png"))
        self.icon.setObjectName("icon")
        self._central_layout.addWidget(self.icon, 0, 0, 1, 1)
        self.title = QtWidgets.QLabel(AboutDialog)
        self.title.setObjectName("title")
        self._central_layout.addWidget(self.title, 0, 1, 1, 1)
        self.information = QtWidgets.QTabWidget(AboutDialog)
        self.information.setObjectName("information")
        self.detailsTab = QtWidgets.QWidget()
        self.detailsTab.setObjectName("detailsTab")
        self._details_layout = QtWidgets.QGridLayout(self.detailsTab)
        self._details_layout.setObjectName("_details_layout")
        self.details = QtWidgets.QLabel(self.detailsTab)
        self.details.setTextInteractionFlags(QtCore.Qt.LinksAccessibleByKeyboard|QtCore.Qt.LinksAccessibleByMouse)
        self.details.setObjectName("details")
        self._details_layout.addWidget(self.details, 0, 0, 1, 1)
        self.information.addTab(self.detailsTab, "")
        self.licenseTab = QtWidgets.QWidget()
        self.licenseTab.setObjectName("licenseTab")
        self._license_layout = QtWidgets.QGridLayout(self.licenseTab)
        self._license_layout.setContentsMargins(0, 0, 0, 0)
        self._license_layout.setObjectName("_license_layout")
        self.license = QtWidgets.QPlainTextEdit(self.licenseTab)
        self.license.setLineWrapMode(QtWidgets.QPlainTextEdit.NoWrap)
        self.license.setReadOnly(True)
        self.license.setObjectName("license")
        self._license_layout.addWidget(self.license, 0, 0, 1, 1)
        self.information.addTab(self.licenseTab, "")
        self._central_layout.addWidget(self.information, 1, 0, 1, 2)
        self.buttons = QtWidgets.QDialogButtonBox(AboutDialog)
        self.buttons.setStandardButtons(QtWidgets.QDialogButtonBox.Ok)
        self.buttons.setObjectName("buttons")
        self._central_layout.addWidget(self.buttons, 2, 0, 1, 2)
        self._central_layout.setColumnStretch(0, 1)
        self._central_layout.setColumnStretch(1, 4)
        self._central_layout.setRowStretch(0, 1)
        self._central_layout.setRowStretch(1, 5)

        self.retranslateUi(AboutDialog)
        self.buttons.accepted.connect(AboutDialog.accept) # type: ignore
        QtCore.QMetaObject.connectSlotsByName(AboutDialog)

    def retranslateUi(self, AboutDialog):
        _translate = QtCore.QCoreApplication.translate
        self.information.setTabText(self.information.indexOf(self.detailsTab), _translate("AboutDialog", "Details"))
        self.information.setTabText(self.information.indexOf(self.licenseTab), _translate("AboutDialog", "License"))
from . import resources_rc



# Copyright (C) 2010-2022, kachelaqa <kachelaqa@gmail.com>

import sys, os, re, glob, copy, errno, socket, json
import urllib.request, urllib.error, http.client
from zipfile import ZipFile, BadZipfile
from html.parser import HTMLParser
from traceback import format_exception
from functools import cmp_to_key
from collections import defaultdict
from multiprocessing import Pool
from threading import Thread
from queue import Queue
from pkgbrowser import alpm, conf, utils
from pkgbrowser.enum import State, Source


AUR_DOM = os.environ.get('AUR_DOM', 'https://aur.archlinux.org')
AUR_RPC = AUR_DOM + '/rpc'
AUR_HTM = AUR_DOM + '/packages'
AUR_SRC = AUR_DOM + '/cgit/aur.git/plain/.SRCINFO'
ALA_DOM = os.environ.get('ALA_DOM', 'https://archive.archlinux.org')
ALA_LIST = ALA_DOM + '/packages/%s/%s/'
ARCH_PKG = 'https://www.archlinux.org/packages/%s/%s/%s'

_arch_repos = set([
    'core', 'extra', 'community', 'multilib',
    'testing', 'community-testing', 'multilib-testing',
    ])


class BackendError(Exception): pass


class DatabaseError(BackendError):
    def __init__(self, source='', errno=None):
        self.source = source
        if errno is None:
            self.errno = alpm.error_number()
        else:
            self.errno = int(errno)
        BackendError.__init__(self, alpm.error_string(self.errno))

    def __str__(self):
        message = self.args[0]
        if self.source:
            message = '%s: %s' % (message, self.source)
        return '[Errno %d] %s' % (self.errno, message)

    def __reduce__(self):
        return (self.__class__, (self.source, self.errno))


class NetworkError(BackendError):
    def __init__(self, url, reason=None):
        if isinstance(reason, socket.timeout):
            message = '[Errno %d] %s' % (
                errno.ETIMEDOUT, os.strerror(errno.ETIMEDOUT))
        elif isinstance(reason, EnvironmentError):
            if reason.errno in errno.errorcode:
                message = '[Errno %d] %s' % (
                    reason.errno, os.strerror(reason.errno))
            else:
                message = reason.strerror
        elif isinstance(reason, int):
            message = '[HTTP %d] %s' % (
                reason, http.client.responses.get(reason, ''))
        elif reason is None:
            message = 'Unknown error'
        else:
            message = str(reason)
        BackendError.__init__(self, message)
        self.url = url

    def __reduce__(self):
        return (self.__class__, (self.url, self.args[0]))


class PatternError(BackendError):
    def __init__(self, text, message):
        BackendError.__init__(self, message)
        self.text = text

    def __reduce__(self):
        return (self.__class__, (self.text, self.args[0]))


class Traceback(Exception):
    @staticmethod
    def format(*args):
        if len(args) == 3:
            return ''.join(format_exception(*args))
        return args and args[0] or ''

    def __init__(self, *args):
        Exception.__init__(self, self.format(*args))

    def __reduce__(self):
        return (self.__class__, self.args)


class Cache(object):
    _caches = {}
    _path = ''
    _offline = False

    @classmethod
    def set_path(cls, path):
        cls._path = path

    @classmethod
    def set_offline(cls, offline):
        cls._offline = bool(offline)

    @classmethod
    def has_files(cls):
        if cls._path:
            return any(os.access(path, os.R_OK) for path in
                       glob.glob(os.path.join(cls._path, '*.files.zip')))
        return False

    @classmethod
    def get_files(cls, package):
        if alpm.pkg_get_installdate(package):
            return alpm.pkg_join_files(package) or ''
        else:
            key = alpm.pkg_get_repository(package)
            cache = cls._caches.get(key)
            if cache is None:
                path = os.path.join(cls._path, '%s.files.zip' % key)
                try:
                    cache = cls._caches[key] = ZipFile(path)
                except (IOError, BadZipfile):
                    pass
            if cache is not None:
                try:
                    files = cache.read(alpm.pkg_get_fullname(package))
                except KeyError:
                    pass
                except (IOError, BadZipfile):
                    cls.clear(key)
                else:
                    return files.decode('utf-8')

    @classmethod
    def get_log(cls, *names):
        key = 'log.zip'
        cache = cls._caches.get(key)
        if cache is None:
            path = alpm.option_get_logfile()
            try:
                cache = cls._caches[key] = conf.load_logfile(path)
            except (IOError, BadZipfile):
                pass
        if cache is not None:
            log = []
            try:
                for name in names:
                    try:
                        entry = cache.read(name)
                    except KeyError:
                        pass
                    else:
                        log.extend(entry.decode('utf-8').splitlines())
                return '\n'.join(sorted(
                    log, key=lambda line: line.partition(']')[0]))
            except (IOError, BadZipfile):
                cls.clear(key)

    @classmethod
    def get_cache(cls, arch, *names):
        if arch == 'any':
            architectures = [os.uname()[4], 'any']
        else:
            architectures = [arch, 'any']
        sources = [[] for name in names]
        key = 'pkgcache.zip'
        cache = cls._caches.get(key)
        if cache is None:
            paths = alpm.option_get_cachedirs()
            try:
                cache = cls._caches[key] = conf.load_pkgcache(paths)
            except (IOError, BadZipfile):
                pass
        if cache is not None:
            try:
                for index, name in enumerate(names):
                    source = sources[index]
                    for arch in architectures:
                        path = '%s/%s' % (name, arch)
                        try:
                            entry = cache.read(path)
                        except KeyError:
                            pass
                        else:
                            source.extend(entry.decode('utf-8').splitlines())
            except (IOError, BadZipfile):
                cls.clear(key)
        if not cls._offline and ALA_DOM:
            parser = AlaParser()
            urls = [utils.make_url(ALA_LIST % (name[0], name))
                    for name in names]
            downloads = Downloader.download(urls, True)
            for index, url in enumerate(urls):
                data = downloads.get(url)
                if data is None:
                    continue
                source = sources[index]
                data = parser.read(data)
                for arch in architectures:
                    path = '%s/%s' % (names[index], arch)
                    if path not in data:
                        continue
                    for package in data[path]:
                        source.append(os.path.join(url, package))
        compare = cmp_to_key(
            lambda a, b, cmp=alpm.pkg_vercmp, match=conf.match_pkgfile:
                cmp(match(b).group(2), match(a).group(2)))
        for source in sources:
            source.sort(key=compare)
        return sources

    @classmethod
    def clear(cls, *args):
        for key in args or list(cls._caches.keys()):
            try:
                cls._caches.pop(key).close()
            except Exception:
                pass


class Downloader(Thread):
    @staticmethod
    def download(urls, quiet=False):
        tasks = Queue()
        output = Queue()
        for index, url in enumerate(urls):
            if index < 10:
                Downloader(tasks, output)
            tasks.put(url)
        tasks.join()
        result = {}
        while not output.empty():
            url, data, exception = output.get()
            if not quiet and exception is not None:
                raise exception
            result[url] = data
        return result

    def __init__(self, tasks, output):
        Thread.__init__(self)
        self._tasks = tasks
        self._output = output
        self.daemon = True
        self.start()

    def run(self):
        while True:
            url = self._tasks.get()
            try:
                response = urllib.request.urlopen(url, timeout=20)
                try:
                    self._output.put((url, response.read(), None))
                finally:
                    response.close()
            except urllib.error.HTTPError as exception:
                self._output.put(
                    (url, None, NetworkError(url, exception.code)))
            except urllib.error.URLError as exception:
                self._output.put(
                    (url, None, NetworkError(url, exception.reason)))
            except BaseException:
                self._output.put(
                    (url, None, Traceback(*sys.exc_info())))
            finally:
                self._tasks.task_done()


class AlaParser(HTMLParser):
    def __init__(self):
        HTMLParser.__init__(self)

    def read(self, data):
        self._link = False
        self._info = defaultdict(list)
        self.reset()
        try:
            self.feed(data.decode('utf-8', 'replace'))
        except AssertionError:
            pass
        return self._info

    def handle_starttag(self, tag, attrs):
        self._link = tag == 'a'

    def handle_data(self, data):
        if self._link:
            data = data.strip()
            match = conf.match_pkgfile(data)
            if match is not None:
                self._info['%s/%s' % match.group(1, 3)].append(data)

    def handle_endtag(self, tag):
        self._link = False


class AurParser(HTMLParser):
    def __init__(self):
        HTMLParser.__init__(self)

    def read(self, data):
        self._key = None
        self._link = False
        self._info = defaultdict(list)
        self.reset()
        try:
            self.feed(data.decode('utf-8', 'replace'))
        except AssertionError:
            pass
        return self._info

    def handle_starttag(self, tag, attrs):
        if tag == 'div':
            if len(self._info) < 2:
                key = dict(attrs).get('id')
                if key == 'pkgreqs':
                    self._key = 'RequiredBy'
        elif tag == 'a' and self._key:
            self._link = True

    def handle_data(self, data):
        if self._link:
            self._info[self._key].append(data.strip())

    def handle_endtag(self, tag):
        if self._link:
            self._link = False
        elif tag == 'div' and self._key:
            self._key = None
            self._link = False


class Matcher(object):
    Not = 1
    Exact = 2
    RegExp = 4
    Group = 8

    def __init__(self, text, getters, files=False):
        self._text = text
        self._getters = getters
        self._targets = self._compile(
            self._parse(iter(text)), files)

    def _parse(self, text, context=0):
        result = []
        element = []
        term = [0, '']
        alt = quoted = False
        char = quote = ''
        while char is not None:
            try:
                if not char:
                    char = next(text)
            except StopIteration:
                char = None
            else:
                if not quote:
                    if char == '(':
                        if not term[1]:
                            term[1] = self._parse(
                                text, term[0] | context | Matcher.Group)
                            if not term[1]:
                                term = [0, '']
                            char = ''
                    elif char == ')':
                        if context & Matcher.Group:
                            if not term[1]:
                                context = 0
                                break
                        else:
                            context |= Matcher.Group
                            break
                    elif char in '"\'':
                        quote = char
                        char = ''
                    elif char in '|~=%':
                        if not term[1]:
                            if char == '~':
                                term[0] |= Matcher.Not
                            elif char == '=':
                                term[0] |= Matcher.Exact
                            elif char == '%':
                                term[0] |= Matcher.RegExp
                            else:
                                alt = True
                            char = ''
                            continue
                    elif not char.isspace():
                        term[1] += char
                        char = ''
                        continue
                    else:
                        char = ''
                elif char == quote:
                    quote = char = ''
                    quoted = True
                else:
                    term[1] += char
                    char = ''
                    continue
            if term[1] or quoted:
                if context & Matcher.Group:
                    if context & Matcher.Not:
                        term[0] ^= Matcher.Not
                        if result:
                            alt = not alt
                    term[0] |= context & (Matcher.Exact | Matcher.RegExp)
                if isinstance(term[1], list):
                    term[0] = Matcher.Group
                if term[0] & Matcher.Group and len(term[1]) == 1:
                    terms = term[1][0]
                else:
                    terms = [term]
                if alt:
                    element = terms
                    alt = False
                else:
                    element.extend(terms)
                if element is not (result and result[-1]):
                    result.append(element)
                term = [0, '']
            quoted = False
        if quote:
            raise PatternError(self._text, 'unclosed quote')
        elif context & Matcher.Group:
            raise PatternError(self._text, 'unclosed parenthesis')
        return result

    def _compile(self, items, files=False):
        result = []
        flags = re.M if files else re.M | re.I
        for item in items:
            element = set()
            for term in item:
                if term[0] & Matcher.Group:
                    term = (term[0], self._compile(term[1], files))
                else:
                    if files and not term[0] & Matcher.RegExp:
                        pattern = '/%s\n' % term[1].lstrip('/')
                        if term[0] & Matcher.Exact:
                            pattern = '\n' + pattern
                        def search(value, pattern=pattern):
                            return pattern in value or None
                    else:
                        if term[0] & Matcher.RegExp:
                            pattern = term[1]
                        elif term[0] & Matcher.Exact:
                            pattern = '^%s$' % re.escape(term[1])
                        else:
                            pattern = '^.*%s.*$' % re.escape(term[1])
                        try:
                            search = re.compile(pattern, flags).search
                        except re.error as exception:
                            raise PatternError(pattern, str(exception))
                    term = (term[0], term[1], search)
                element.add(term)
            element = tuple(sorted(element, key=cmp_to_key(self._compare)))
            if element not in result:
                result.append(element)
        return tuple(result)

    def _compare(self, a, b):
        if a[0] & (Matcher.Not | Matcher.RegExp):
            if not b[0] & (Matcher.Not | Matcher.RegExp):
                return 1
        elif b[0] & (Matcher.Not | Matcher.RegExp):
            return -1
        if isinstance(a[1], (tuple, list)):
            if isinstance(b[1], (tuple, list)):
                result = self._compare(a[1][0][0], b[1][0][0])
                if not result:
                    return len(a[1]) - len(b[1])
                return result
            return 1
        elif isinstance(b[1], (tuple, list)):
            return -1
        if a[0] & Matcher.Exact:
            if not b[0] & Matcher.Exact:
                return -1
        elif b[0] & Matcher.Exact:
            return 1
        return len(b[1]) - len(a[1])

    def prioritize(self, strict=False):
        return self._prioritize(self._targets, strict)

    def _prioritize(self, items, strict=False):
        result = []
        for index, element in enumerate(items):
            term = element[0]
            if strict and term[0] & Matcher.Not:
                raise PatternError(self._text, 'unqualified negative term')
            if strict and term[0] & Matcher.RegExp:
                raise PatternError(self._text, 'unqualified regexp term')
            if term[0] & Matcher.Group:
                result.extend(self._prioritize(term[1], strict))
            else:
                result.append((bool(term[0] & Matcher.Exact), term[1]))
        return result

    def match(self, item):
        return self._match(item, self._targets)

    def _match(self, item, targets):
        for element in targets:
            terms = set(element)
            found = set()
            for getter in self._getters:
                value = getter(item) or ''
                for term in (terms - found):
                    if term[0] & Matcher.Group:
                        match = self._match(item, term[1])
                    else:
                        match = term[2](value) is not None
                        if term[0] & Matcher.Not:
                            match = not match
                    if match:
                        found.add(term)
                if found == terms:
                    return True
        return False


def _call(args):
    return getattr(backend, args[0])(*args[1:])

def _call_async(args):
    try:
        return _call(args)
    except BackendError as exception:
        return exception
    except BaseException:
        return Traceback(*sys.exc_info())


class Backend(object):
    def __init__(self):
        self._aurparser = AurParser()
        self._rpcs = {}
        self._pool = None
        self._callback = None
        self._offline = False

    def version(self):
        return alpm.version()

    def initialize(self):
        try:
            config = conf.read_config()
        except IOError as exception:
            raise DatabaseError(exception.filename, alpm.ERR_NOT_A_FILE)
        else:
            self.release()
            rootdir = config.get('RootDir')
            if rootdir:
                dbpath = os.path.join(rootdir, conf.PM_DB_PATH)
            else:
                rootdir = conf.PM_ROOT_DIR
                dbpath = config.get('DBPath', conf.PM_DB_PATH)
            error = alpm.initialize(rootdir, dbpath)
            if error == alpm.ERR_NOT_A_DIR:
                if not os.path.isdir(rootdir):
                    raise DatabaseError(rootdir, error)
                raise DatabaseError(dbpath, error)
            elif error:
                raise DatabaseError(None, error)
            for arch in config.get('Architecture', ()):
                alpm.option_set_arch(arch)
            for name in config['Repositories']:
                if alpm.register_syncdb(name) is None:
                    raise DatabaseError(name)
            alpm.option_set_logfile(config.get('LogFile', conf.PM_LOG_FILE))
            for path in config.get('CacheDir', conf.PM_CACHE_DIRS):
                alpm.option_add_cachedir(path)
            if self._callback is not None:
                _callback = self._callback
                def callback(items, exception):
                    if exception is None:
                        self._rpcs = dict(items)
                    self._callback = _callback
                    _callback([], exception)
                self.set_callback(callback)
                self._call([['_load']])
            else:
                self._rpcs = dict(self._load())

    def _load(self):
        items = []
        if alpm.is_initialized():
            targets = defaultdict(list)
            source = Source.Local | Source.Foreign
            for location, package in self._iter_packages(source):
                targets[('info', '')].append(alpm.pkg_get_name(package))
            for info in self._fetch_packages(targets):
                items.append((info['Name'], (
                    info['PackageBase'],
                    info['Version'],
                    info['Maintainer'] or '',
                    info['NumVotes'],
                    info['Popularity'],
                    )))
        return items

    def release(self):
        Cache.clear()
        self._rpcs.clear()
        if alpm.is_initialized() and alpm.release() != 0:
            raise DatabaseError()

    def set_offline(self, offline):
        self._offline = bool(offline)
        Cache.set_offline(offline)

    def set_callback(self, callback):
        if self._pool is not None:
            self._pool.terminate()
        self._callback = callback

    def _call(self, args):
        self._pool = Pool()
        if self._callback is not None:
            def callback(results):
                items = []
                exception = None
                for result in results:
                    if not isinstance(result, BaseException):
                        items.extend(result)
                    else:
                        exception = result
                        items = None
                        break
                self._callback(items, exception)
            self._pool.map_async(_call_async, args, callback=callback)
        else:
            items = []
            for item in self._pool.map(_call, args):
                items.extend(item)
            self._pool.terminate()
            return items

    def _iter_dbs(self, source=0, locations=()):
        if not source:
            source = Source.Sync | Source.Local
        if source & Source.Sync:
            item = alpm.get_syncdbs()
            while item is not None:
                db = alpm.list_get_db(item)
                name = alpm.db_get_name(db)
                if not locations or name in locations:
                    yield name, db
                item = alpm.list_next(item)
        if source & Source.Local:
            db = alpm.get_localdb()
            name = alpm.db_get_name(db)
            if not locations or name in locations:
                yield name, db

    def _iter_groups(self, locations=(), match=None):
        for location, db in self._iter_dbs(Source.Sync, locations):
            item = alpm.db_get_groupcache(db)
            while item is not None:
                group = alpm.list_get_group(item)
                item = alpm.list_next(item)
                if match and not match(group):
                    continue
                yield location, group

    def _iter_group(self, locations=(), targets=()):
        for location, db in self._iter_dbs(Source.Sync, locations):
            for target in targets or ():
                group = alpm.db_get_group(db, target)
                item = alpm.group_get_pkgs(group)
                while item is not None:
                    yield location, alpm.list_get_pkg(item)
                    item = alpm.list_next(item)

    def _iter_packages(self, source=0, locations=(), match=None):
        local = alpm.db_get_name(alpm.get_localdb())
        for location, db in self._iter_dbs(source, locations):
            item = alpm.db_get_pkgcache(db)
            while item is not None:
                package = alpm.list_get_pkg(item)
                item = alpm.list_next(item)
                if ((location == local and source & Source.Foreign and
                     alpm.pkg_get_repository(package) != local) or
                    (match and not match(package))):
                    continue
                yield location, package

    def _dispatch(self, keys, source=0):
        if source == Source.Group:
            dispatch = {
                'name': alpm.group_get_name,
                }
        elif source == Source.AUR:
            dispatch = {
                'name': lambda item: item['Name'],
                'description': lambda item: item['Description'],
                'maintainer': lambda item: item['Maintainer'],
                }
        else:
            def pkg_get_maintainer(item):
                name = alpm.pkg_get_name(item)
                if name in self._rpcs:
                    return self._rpcs[name][2]
                return alpm.pkg_get_packager(item)
            dispatch = {
                'name': alpm.pkg_get_name,
                'description': alpm.pkg_get_desc,
                'provides': alpm.pkg_join_provides,
                'replaces': alpm.pkg_join_replaces,
                'depends': alpm.pkg_join_depends,
                'optdepends': alpm.pkg_join_optdepends,
                'maintainer': pkg_get_maintainer,
                'files': Cache.get_files,
                }
        return tuple(dispatch[key] for key in keys if key in dispatch)

    def _fetch_packages(self, targets):
        packages = []
        if not self._offline:
            urls = []
            for (mode, by), targets in targets.items():
                last = len(targets)
                defaults = dict(type=mode, v=5)
                if mode == 'info':
                    key, limit = 'arg[]', 300
                else:
                    key, limit = 'arg', 1
                    defaults['by'] = by
                query = defaultdict(list, defaults)
                for index, target in enumerate(targets, 1):
                    query[key].append(target)
                    if not index % limit or index == last:
                        urls.append(utils.make_url(AUR_RPC, query))
                        query = defaultdict(list, defaults)
            for url, data in Downloader.download(urls).items():
                data = json.loads(data.decode('utf-8', 'replace'))
                if data['type'] == 'error':
                    raise NetworkError(url, data['error'])
                elif data['resultcount']:
                    packages.extend(data['results'])
        return packages

    def _fetch_package(self, name, basename=None, update=False):
        if not self._offline:
            info = None
            urls = {
                'rpc': utils.make_url(
                    AUR_RPC, dict(type='info', arg=name, v=5)),
                'htm': utils.make_url(AUR_HTM, dict(N=name)),
                }
            while True:
                if basename:
                    urls['src'] = utils.make_url(AUR_SRC, dict(h=basename))
                downloads = Downloader.download(urls.values())
                if 'rpc' in urls:
                    url = urls.pop('rpc')
                    data = downloads.get(url)
                    if data is not None:
                        data = json.loads(data.decode('utf-8', 'replace'))
                        if data['type'] == 'error':
                            raise NetworkError(url, data['error'])
                        elif data['resultcount']:
                            info = data['results'][0]
                            data = downloads.get(urls.pop('htm'))
                            if data is not None:
                                info.update(self._aurparser.read(data))
                            if not basename:
                                basename = info['PackageBase']
                                continue
                if info is not None:
                    data = downloads.get(urls['src'])
                    if data is not None:
                        info.update(conf.load_srcinfo(name, data))
                    return AurPackage(info, update=update)
                break
        return NullPackage(name)

    def get_package(self, target, location=None, state=State.Unknown):
        if isinstance(target, Summary):
            state = target.state
            if state & State.AUR and not state & State.Installed:
                location = target.basename
            else:
                location = target.repository
            target = target.name
        update = bool(state & State.Update)
        if state & State.AUR and not state & State.Installed:
            return self._fetch_package(target, location, update)
        elif not state & State.Group:
            locations = location and [location]
            local = localdb = None
            if state & State.Installed or state & State.Unknown:
                localdb = alpm.get_localdb()
                local = alpm.db_get_pkg(localdb, target)
            if state & State.NonInstalled or state & State.Unknown:
                for location, db in self._iter_dbs(Source.Sync, locations):
                    sync = alpm.db_get_pkg(db, target)
                    if sync is not None:
                        if state & State.NonInstalled or local is None:
                            return Package(sync, update=update)
                        else:
                            return Package(local)
                    elif locations:
                        return NullPackage(target)
            if (local is not None and
               (not locations or state & State.Installed)):
                if ((state & State.Foreign or state & State.Unknown) and
                    not state & State.Database):
                    rpc = self._rpcs.get(target)
                    if rpc is not None:
                        aur = self._fetch_package(target, rpc[0], update)
                        return Package(local, aur)
                return Package(local)
            elif not locations and state & State.Unknown:
                provider = alpm.db_find_provider(target, 0)
                if provider is not None:
                    return Package(provider)
                replacer = alpm.db_find_replacer(target, 0)
                if replacer is not None:
                    return Package(replacer)
                if not state & State.Database:
                    return self._fetch_package(target, None, update)
        return NullPackage(target)

    def find(self, text, filters=0, keys=()):
        keys = keys or ['name']
        args = []
        if filters & State.AUR and filters & State.NonInstalled:
            args.append(('_find_aur', text, filters, keys))
        for location in self.list_repositories():
            args.append(('_find', text, filters, keys, [location]))
        return self._call(args)

    def _find(self, text, filters=0, keys=(), locations=()):
        items = []
        if filters & State.Group:
            filters &= ~State.Group
            matcher = Matcher(text, self._dispatch(keys, Source.Group))
            iterator = self._iter_groups(locations, matcher.match)
            items.extend(self._filter_groups(iterator, filters))
        else:
            filters &= ~State.AUR
            args = defaultdict(list)
            for key in keys:
                if key == 'files' and not filters & State.NonInstalled:
                    source = Source.Local
                else:
                    source = Source.Sync | Source.Local | Source.Foreign
                args[source].append(key)
            for source, keys in args.items():
                matcher = Matcher(text, self._dispatch(keys), 'files' in keys)
                iterator = self._iter_packages(
                    source, locations, matcher.match)
                items.extend(self._filter_packages(iterator, filters))
        return items

    def _find_aur(self, text, filters=0, keys=()):
        matcher = Matcher(text, self._dispatch(keys, Source.AUR))
        terms = matcher.prioritize(True)
        maintainer = 'maintainer' in keys
        if 'name' in keys and 'description' not in keys:
            by = 'name'
        else:
            by = 'name-desc'
        targets = defaultdict(list)
        for exact, target in terms:
            if maintainer:
                targets[('search', 'maintainer')].append(target)
            if not maintainer or len(keys) > 1:
                if exact:
                    targets[('info', '')].append(target)
                else:
                    targets[('search', by)].append(target)
        items = []
        seen = set()
        for package in self._fetch_packages(targets):
            identifier = package['ID']
            if identifier not in seen:
                seen.add(identifier)
                if matcher.match(package):
                    items.append(package)
        return self._filter_aur(items, filters)

    def _filter_aur(self, items, filters=0):
        output = []
        for package in items:
            name = package['Name']
            if name not in self._rpcs:
                summary = Summary()
                summary.name = name
                summary.version = package['Version']
                summary.repository = 'aur'
                summary.state = State.NonInstalled | State.AUR
                summary.basename = package['PackageBase']
                summary.votes = package['NumVotes']
                summary.popularity = package['Popularity']
                output.append(summary)
        return output

    def _filter_groups(self, items, filters=0):
        output = []
        if not filters:
            filters = State.Installed | State.NonInstalled | State.Update
        for repository, group in items:
            name = alpm.group_get_name(group)
            summary = Summary()
            summary.name = name
            summary.repository = repository
            summary.state = State.Group | State.Installed
            locations = [repository] if repository else None
            for location, package in self._iter_group(locations, [name]):
                if (summary.state & State.Installed and
                    alpm.pkg_get_status(package) & State.NonInstalled):
                    summary.state = State.Group | State.NonInstalled
                summary.size += alpm.pkg_get_isize(package)
            if filters & summary.state:
                output.append(summary)
        return output

    def _filter_packages(self, items, filters=0):
        output = []
        if not filters:
            filters = State.Installed | State.NonInstalled | State.Update
        localdb = alpm.get_localdb()
        local = alpm.db_get_name(localdb)
        for repository, package in items:
            update = None
            summary = Summary()
            summary.repository = repository
            summary.name = alpm.pkg_get_name(package)
            summary.version = alpm.pkg_get_version(package)
            summary.state = alpm.pkg_get_status(package)
            summary.size = alpm.pkg_get_isize(package)
            if summary.state & State.Installed:
                if summary.repository in (local, 'aur'):
                    rpc = self._rpcs.get(summary.name)
                    if rpc is not None:
                        summary.repository = 'aur'
                        summary.basename = rpc[0]
                        summary.version = rpc[1]
                        summary.state |= State.AUR
                        summary.votes = rpc[3]
                        summary.popularity = rpc[4]
                    elif summary.repository == local:
                        summary.repository = alpm.pkg_get_repository(package)
                if filters & State.Update:
                    state = alpm.pkg_check_update(summary.name,
                                                  summary.version)
                    if filters & state:
                        update = copy.copy(summary)
                        update.state = state | summary.state & State.AUR
                        if update.state & State.AUR:
                            update.size = -1
                        output.append(update)
                current = alpm.db_get_pkg(localdb, summary.name)
                summary.version = alpm.pkg_get_version(current)
                summary.date = alpm.pkg_get_installdate(current)
            if (filters & summary.state and (
                update is None or summary.state & State.AUR or
                summary.repository == alpm.pkg_get_repository(current))):
                output.append(summary)
        return output

    def list_targets(self, targets):
        return self._call([['_list_targets', targets]])

    def _list_targets(self, targets):
        unknown = []
        packages = []
        for target in set(targets):
            package = self.get_package(
                target, state=(State.Database | State.Unknown))
            if package['state'] & State.Unknown:
                unknown.append(target)
            else:
                packages.append((package['repository'], package.base()))
        items = self._filter_packages(packages)
        items.extend(self._filter_aur(
            self._fetch_packages({('info', ''): unknown})))
        return items

    def list_packages(self, filters=0, location=None):
        if filters and filters & State.Foreign:
            source = Source.Local | Source.Foreign
            filters &= ~State.Foreign
        elif location:
            source = Source.Sync
        else:
            source = Source.Sync | Source.Local | Source.Foreign
        packages = self._iter_packages(source, location and [location])
        return self._filter_packages(packages, filters)

    def list_group(self, location=None, target=None):
        packages = self._iter_group(location and [location],
                                    target and [target])
        return self._filter_packages(packages)

    def list_groups(self, location=None):
        groups = self._iter_groups(location and [location])
        return self._filter_groups(groups)

    def list_repositories(self):
        return [repository for repository, db in self._iter_dbs()]

    def statistics(self):
        data = defaultdict(list)
        for package in self.list_packages(State.Installed):
            for key in ('total', package.repository):
                item = data[key]
                if not item:
                    item[:] = [0, 0]
                item[0] += 1
                item[1] += package.size
        stats = []
        for key in self.list_repositories() + ['aur', 'total']:
            item = data.get(key)
            if item is not None:
                stats.append((key, item[0], item[1]))
        return stats


class Summary(object):
    name = ''
    version = ''
    repository = ''
    basename = ''
    state = 0
    date = 0
    size = -1
    votes = -1
    popularity = -1


class BasePackage(object):
    _base = None
    _match = re.compile(r'^([^\s<>=:]*)\s*(.*?)$').match

    def _tree(self):
        result = {'installed': 0, 'missing': 0, 'aur': 0,
                  'isize': 0, 'msize': 0}
        def tree(parent, seen=None):
            output = []
            if seen is None:
                seen = set()
            for depends in (parent.get('depends', ()),
                            parent.get('makedepends', ())):
                for provides, data in depends:
                    if provides in seen:
                        continue
                    seen.add(provides)
                    package = backend.get_package(provides)
                    if isinstance(package, NullPackage):
                        continue
                    name = package['name']
                    state = package['state']
                    if (name == provides and
                        not state & (State.Installed | State.AUR)):
                        provider = alpm.db_find_provider(provides, 1)
                        if provider is not None:
                            package = Package(provider)
                            name = package['name']
                            state = package['state']
                    if name != provides:
                        if name in seen:
                            continue
                        seen.add(name)
                    else:
                        provides = None
                    depends = tree(package, seen)
                    output.append((name, provides, state, depends))
                    if state & State.Installed:
                        result['isize'] += package['size']
                        result['installed'] += 1
                    elif not state & State.AUR:
                        result['msize'] += package['size']
                        result['missing'] += 1
                    else:
                        result['aur'] += 1
            return output
        result['packages'] = tree(self)
        return result

    def _log(self):
        names = (self['replaces'] or '').split()
        names.append(self['name'])
        return Cache.get_log(*names)

    def _cache(self):
        names = [self['name']]
        names.extend((self['replaces'] or '').split())
        return Cache.get_cache(self['arch'], *names)

    def _backup(self):
        item = alpm.pkg_get_backup(self._base)
        while item is not None:
            backup = alpm.list_get_backup(item)
            if backup is not None:
                yield backup
            item = alpm.list_next(item)

    def get(self, key, default=None):
        try:
            return self.__getitem__(key)
        except KeyError:
            return default

    def base(self):
        return self._base


class Package(BasePackage):
    def __init__(self, base, aur=None, update=False):
        self._base = base
        self._aur = aur
        self._update = update

    def _map(self, node):
        items = []
        while node is not None:
            items.append(self._match(alpm.list_get_str(node)).groups())
            node = alpm.list_next(node)
        return items

    def __getitem__(self, key):
        if key == 'name':
            return alpm.pkg_get_name(self._base)
        elif key == 'version':
            return alpm.pkg_get_version(self._base)
        elif key == 'description':
            return alpm.pkg_get_desc(self._base)
        elif key == 'state':
            if self._update:
                state = alpm.pkg_check_update(self['name'], self['version'])
            else:
                state = alpm.pkg_get_status(self._base)
            if isinstance(self._aur, AurPackage):
                state |= State.AUR
            return state
        elif key == 'url':
            return alpm.pkg_get_url(self._base)
        elif key == 'pkgurl':
            repository = self['repository']
            if repository in _arch_repos:
                return ARCH_PKG % (repository, self['arch'], self['name'])
        elif key == 'license':
            licenses = alpm.pkg_get_licenses(self._base)
            return alpm.list_join_str(licenses, ', ')
        elif key == 'repository':
            if isinstance(self._aur, AurPackage):
                return 'aur'
            return alpm.pkg_get_repository(self._base)
        elif key == 'groups':
            groups = alpm.pkg_get_groups(self._base)
            return alpm.list_join_str(groups, ' ')
        elif key == 'provides':
            provides = alpm.pkg_get_provides(self._base)
            return alpm.list_join_str(provides, ' ')
        elif key == 'depends':
            depends = []
            node = alpm.pkg_get_depends(self._base)
            while node is not None:
                data = alpm.dep_compute_string(alpm.list_get_dep(node))
                depends.append(self._match(data).groups())
                node = alpm.list_next(node)
            return depends
        elif key == 'optdepends':
            optdepends = []
            node = alpm.pkg_get_optdepends(self._base)
            while node is not None:
                data = alpm.dep_compute_string(alpm.list_get_dep(node))
                optdepends.append(self._match(data).groups())
                node = alpm.list_next(node)
            return optdepends
        elif key == 'conflicts':
            return self._map(alpm.pkg_get_conflicts(self._base))
        elif key == 'replaces':
            replaces = alpm.pkg_get_replaces(self._base)
            return alpm.list_join_str(replaces, ' ')
        elif key == 'required':
            required = self._map(alpm.pkg_compute_requiredby(self._base))
            if isinstance(self._aur, AurPackage):
                required.extend(self._aur[key])
            return required
        elif key == 'optional':
            return self._map(alpm.pkg_compute_optionalfor(self._base))
        elif key == 'validation':
            return alpm.pkg_get_validation(self._base)
        elif key == 'installed':
            return alpm.pkg_get_installdate(self._base)
        elif key == 'download':
            return alpm.pkg_get_size(self._base)
        elif key == 'size':
            return alpm.pkg_get_isize(self._base)
        elif key == 'packager':
            return alpm.pkg_get_packager(self._base)
        elif key == 'arch':
            return alpm.pkg_get_arch(self._base)
        elif key == 'script':
            return alpm.pkg_has_scriptlet(self._base)
        elif key == 'built':
            return alpm.pkg_get_builddate(self._base)
        elif key == 'files':
            return Cache.get_files(self._base)
        elif key == 'backup':
            return self._backup()
        elif key == 'tree':
            return self._tree()
        elif key == 'log':
            return self._log()
        elif key == 'cache':
            return self._cache()
        elif isinstance(self._aur, AurPackage):
            return self._aur.get(key)
        else:
            raise KeyError(key)


class AurPackage(BasePackage):
    def __init__(self, base, update=False):
        self._base = base
        self._update = update

    def _map(self, strings):
        if strings:
            if not isinstance(strings, list):
                strings = [strings]
            return [self._match(string).groups() for string in strings]
        return []

    def _list(self, strings, sep=' '):
        if isinstance(strings, list):
            strings = sep.join(strings)
        return strings

    def _find(self, key, alt):
        value = self._base.get(alt)
        if not value or not value[0]:
            value = self._base.get(key)
        return value

    def __getitem__(self, key):
        if key == 'name':
            name = self._find(key, 'Name')
            if isinstance(name, list):
                return name[0]
            return name
        elif key == 'version':
            return self._find(key, 'Version')
        elif key == 'description':
            return self._find(key, 'Description')
        elif key == 'state':
            state = State.AUR
            if self._update:
                state |= alpm.pkg_check_update(self['name'], self['version'])
            return state
        elif key == 'url':
            return self._find(key, 'URL')
        elif key == 'aururl':
            try:
                return '%s?ID=%d' % (AUR_HTM, int(self._base.get('ID')))
            except (ValueError, TypeError):
                pass
        elif key == 'license':
            license = self._find(key, 'License')
            if isinstance(license, str):
                license = license.split()
            return self._list(license, ', ')
        elif key == 'outdated':
            try:
                return int(self._base.get('OutOfDate'))
            except (ValueError, TypeError):
                pass
        elif key == 'votes':
            try:
                return int(self._base.get('NumVotes'))
            except (ValueError, TypeError):
                pass
        elif key == 'popularity':
            try:
                return float(self._base.get('Popularity'))
            except (ValueError, TypeError):
                pass
        elif key == 'repository':
            return 'aur'
        elif key == 'installed':
            return None
        elif key == 'files':
            return None
        elif key == 'backup':
            return None
        elif key == 'tree':
            return self._tree()
        elif key == 'log':
            return self._log()
        elif key == 'cache':
            return self._cache()
        elif key == 'arch':
            return self._list(self._base.get('arch'), ', ')
        elif key == 'groups':
            return self._list(self._find(key, 'Groups'))
        elif key == 'provides':
            return self._list(self._find(key, 'Provides'))
        elif key == 'depends':
            return self._map(self._find(key, 'Depends'))
        elif key == 'makedepends':
            return self._map(self._find(key, 'MakeDepends'))
        elif key == 'optdepends':
            return self._map(self._find(key, 'OptDepends'))
        elif key == 'conflicts':
            return self._map(self._find(key, 'Conflicts'))
        elif key == 'replaces':
            return self._list(self._find(key, 'Replaces'))
        elif key == 'required':
            return self._map(self._base.get('RequiredBy'))
        elif key == 'submitted':
            try:
                return int(self._base.get('FirstSubmitted'))
            except (ValueError, TypeError):
                pass
        elif key == 'modified':
            try:
                return int(self._base.get('LastModified'))
            except (ValueError, TypeError):
                pass
        elif key == 'maintainer':
            return self._base.get('Maintainer')
        else:
            return self._base[key]


class NullPackage(BasePackage):
    def __init__(self, name):
        self._name = name

    def __getitem__(self, key):
        if key == 'name':
            return self._name
        elif key == 'state':
            return State.Unknown


backend = Backend()



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'designer/help.ui'
#
# Created by: PyQt5 UI code generator 5.15.6
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_HelpDialog(object):
    def setupUi(self, HelpDialog):
        HelpDialog.setObjectName("HelpDialog")
        HelpDialog.resize(760, 580)
        icon = QtGui.QIcon()
        icon.addPixmap(QtGui.QPixmap(":/icons/app.png"), QtGui.QIcon.Normal, QtGui.QIcon.Off)
        HelpDialog.setWindowIcon(icon)
        self.gridLayout = QtWidgets.QGridLayout(HelpDialog)
        self.gridLayout.setObjectName("gridLayout")
        self.browser = QtWidgets.QTextBrowser(HelpDialog)
        self.browser.setOpenLinks(False)
        self.browser.setObjectName("browser")
        self.gridLayout.addWidget(self.browser, 0, 0, 1, 1)
        self.buttons = QtWidgets.QDialogButtonBox(HelpDialog)
        self.buttons.setStandardButtons(QtWidgets.QDialogButtonBox.Close)
        self.buttons.setObjectName("buttons")
        self.gridLayout.addWidget(self.buttons, 1, 0, 1, 1)

        self.retranslateUi(HelpDialog)
        self.buttons.rejected.connect(HelpDialog.reject) # type: ignore
        QtCore.QMetaObject.connectSlotsByName(HelpDialog)

    def retranslateUi(self, HelpDialog):
        pass
from . import resources_rc



# Copyright (C) 2010-2022, kachelaqa <kachelaqa@gmail.com>

import sys, time
from PyQt5.QtCore import (
    Qt, QObject, QTimer, QEvent, QSignalMapper, QFile, QDir, QUrl,
    QTextStream, QStringListModel,
    )
from PyQt5.QtGui import (
    QInputEvent, QKeySequence, QIcon, QTextCursor, QTextDocument,
    QStandardItemModel, QStandardItem, QFontMetrics, QFont, QPalette,
    )
from PyQt5.QtWidgets import (
    qApp, QMainWindow, QDialog, QMessageBox, QFileDialog, QHBoxLayout,
    QWidget, QProgressBar, QShortcut, QAction, QMenu, QButtonGroup,
    QGroupBox, QToolButton, QRadioButton, QCheckBox, QTextBrowser,
    QTreeWidgetItem, QHeaderView,
    )
from pkgbrowser.backend import (
    backend, DatabaseError, NetworkError, PatternError, Traceback, Cache,
    )
from pkgbrowser.fmt import Format
from pkgbrowser.enum import State
from pkgbrowser.ui.window import Ui_Window
from pkgbrowser.ui.about import Ui_AboutDialog
from pkgbrowser.ui.help import Ui_HelpDialog


def delayed(method):
    def wrapper(*args, **kwargs):
        QTimer.singleShot(5, lambda: method(*args, **kwargs))
    return wrapper


class Window(QMainWindow, Ui_Window):
    def __init__(self):
        QMainWindow.__init__(self)
        self.setupUi(self)
        self._aborting = 0
        self._active = False
        self._help = None
        self._about = None
        self._dialog = None
        self._package = None
        self._history = []
        self._bookmarks = []
        self._index = 0
        self.setWindowTitle(qApp.applicationTitle())
        self.iconPackage = QIcon(':/icons/package.png')
        self.iconGroup = QIcon(':/icons/group.png')
        self.iconRepository = QIcon(':/icons/repository.png')
        self.iconFilter = QIcon(':/icons/filter.png')
        self.iconSearch = QIcon(':/icons/search.png')
        self.iconStop = QIcon(':/icons/stop.png')
        self.iconOrphan = QIcon(':/icons/orphan.png')
        self.iconInstalled = QIcon(':/icons/installed.png')
        self.iconDependency = QIcon(':/icons/dependency.png')
        self.iconOptional = QIcon(':/icons/optional.png')
        self.iconUpgrade = QIcon(':/icons/upgrade.png')
        self.iconDowngrade = QIcon(':/icons/downgrade.png')
        self.format = Formatter(self)
        self.centralWidget().installEventFilter(self)
        self.menuBar().installEventFilter(self)
        self.progress = QProgressBar()
        self.progress.setFixedSize(150, 16)
        self.statusBar().addPermanentWidget(self.progress)
        self.progress.hide()
        model = QStandardItemModel(self.packages)
        self.packages.setModel(model)
        header = self.packages.header()
        header.setSectionsMovable(False)
        header.setSectionsClickable(True)
        header.setSectionResizeMode(QHeaderView.ResizeToContents)
        header.setContextMenuPolicy(Qt.CustomContextMenu)
        header.customContextMenuRequested.connect(self.handleHeaderMenu)
        for index, (text, numeric, order, hidden) in enumerate((
            (self.tr('Package'), False, Qt.AscendingOrder, False),
            (self.tr('Version'), False, Qt.AscendingOrder, False),
            (self.tr('Repository'), False, Qt.AscendingOrder, False),
            (self.tr('Status'), False, Qt.AscendingOrder, False),
            (self.tr('Date'), True, Qt.DescendingOrder, False),
            (self.tr('Size'), True, Qt.AscendingOrder, False),
            (self.tr('Votes'), True, Qt.DescendingOrder, True),
            (self.tr('Popularity'), True, Qt.DescendingOrder, True),
            )):
            item = QStandardItem(text)
            item.setData(order, Qt.InitialSortOrderRole)
            item.setData(Qt.UserRole if numeric else Qt.DisplayRole)
            model.setHorizontalHeaderItem(index, item)
            if index == 0:
                header.setSortIndicator(index, order)
            header.setSectionHidden(index, hidden)
        self.filters.installEventFilter(self)
        self.filters.viewport().installEventFilter(self)
        corner = QWidget(self.information)
        self.backButton = QToolButton(corner)
        self.backButton.setFocusPolicy(Qt.NoFocus)
        self.backButton.setArrowType(Qt.LeftArrow)
        self.forwardButton = QToolButton(corner)
        self.forwardButton.setFocusPolicy(Qt.NoFocus)
        self.forwardButton.setArrowType(Qt.RightArrow)
        self.bookmarkButton = QToolButton(corner)
        self.bookmarkButton.setFocusPolicy(Qt.NoFocus)
        self.bookmarkButton.setIcon(QIcon(':/icons/bookmark.png'))
        self.bookmarkButton.setPopupMode(QToolButton.MenuButtonPopup)
        self.bookmarkMenu = QMenu(self)
        self.bookmarkMenu.installEventFilter(self)
        self.bookmarkButton.setMenu(self.bookmarkMenu)
        layout = QHBoxLayout(corner)
        layout.setContentsMargins(0, 0, 0, 1)
        layout.setSpacing(0)
        layout.addWidget(self.backButton)
        layout.addWidget(self.forwardButton)
        layout.addWidget(self.bookmarkButton)
        self.information.setCornerWidget(corner)
        self.information.installEventFilter(self)
        self.searchBox.lineEdit().setPlaceholderText(self.tr('Search'))
        self.searchBox.setCompleter(None)
        self.searchBox.view().installEventFilter(self)
        self.searchBox.setModel(QStringListModel(self.searchBox))
        self.searchFiles.setHidden(True)
        self.searchButton.setEnabled(False)
        self.keyGroup.installEventFilter(self)
        self.scopeGroup.installEventFilter(self)
        self.filterButtons = QButtonGroup(self.searchOptions)
        self.filterButtons.addButton(self.scopeAll, 0)
        self.filterButtons.addButton(self.scopeInstalled, 1)
        self.filterButtons.addButton(self.scopeNonInstalled, 2)
        self.typeButtons = QButtonGroup(self.scopeGroup)
        self.typeButtons.addButton(self.scopePackages, 0)
        self.typeButtons.addButton(self.scopeGroups, 1)
        self.otherButtons = QSignalMapper(self.searchOptions)
        for button in (
            self.keyNames, self.keyDescriptions,
            self.keyDepends, self.keyProvides,
            self.keyReplaces, self.keyOptional,
            self.keyMaintainer, self.keyFiles,
            self.scopeAUR,
            ):
            self.otherButtons.setMapping(button, button)
            button.clicked.connect(self.otherButtons.map)
        self.keyNames.setChecked(True)
        self.keyDescriptions.setChecked(True)
        self.scopeAll.setChecked(True)
        self.scopePackages.setChecked(True)
        self.findBar.setVisible(False)
        self.showFindBar = QAction(self.tr('Find...'), self)
        self.addAction(self.showFindBar)
        self.fileRefresh.setShortcut(QKeySequence(self.tr('F5')))
        self.fileCancel.setShortcut(QKeySequence(self.tr('Ctrl+W')))
        self.fileQuit.setShortcut(QKeySequence(self.tr('Ctrl+Q')))
        self.toolsStatistics.setShortcut(QKeySequence(self.tr('Ctrl+S')))
        self.toolsCopy.setShortcut(QKeySequence(self.tr('Ctrl+L')))
        self.helpManual.setShortcut(QKeySequence(self.tr('F1')))
        self.searchButton.setShortcut(QKeySequence(self.tr('F3')))
        self.searchFiles.setShortcut(QKeySequence(self.tr('Ctrl+O')))
        self.backButton.setShortcut(QKeySequence(self.tr('Alt+Left')))
        self.forwardButton.setShortcut(QKeySequence(self.tr('Alt+Right')))
        self.bookmarkButton.setShortcut(QKeySequence(self.tr('Ctrl+B')))
        self.showFindBar.setShortcut(QKeySequence(self.tr('Ctrl+F')))
        shortcut = QShortcut(QKeySequence(self.tr('Escape')),
            self.findBar, self.findBar.hide)
        shortcut.setContext(Qt.WidgetWithChildrenShortcut)
        QShortcut(QKeySequence(self.tr('F7')),
            self, self.bookmarkButton.showMenu)
        QShortcut(QKeySequence(self.tr('F9')),
            self, self.filters.setFocus)
        QShortcut(QKeySequence(self.tr('F10')),
            self, self.packages.setFocus)
        QShortcut(QKeySequence(self.tr('F11')),
            self, self.information.setFocus)
        QShortcut(QKeySequence(self.tr('F12')),
            self, self.activateSearchBox)
        self.fileOffline.triggered.connect(self.handleWorkOffline)
        self.fileRefresh.triggered.connect(self.handleRefresh)
        self.fileCancel.triggered.connect(self.handleCancelTask)
        self.fileQuit.triggered.connect(self.close)
        self.toolsStatistics.triggered.connect(
            lambda: self.handleStatistics())
        self.toolsCopy.triggered.connect(self.handleCopyList)
        self.toolsClear.triggered.connect(self.handleClearBookmarks)
        self.helpManual.triggered.connect(lambda: self.handleManual())
        self.helpAbout.triggered.connect(lambda: self.handleAbout())
        self.helpAboutQt.triggered.connect(lambda: self.handleAboutQt())
        self.filters.itemSelectionChanged.connect(self.handleFilterActivated)
        self.filters.itemExpanded.connect(self.handleFilterExpanded)
        self.searchBox.editTextChanged.connect(self.handleSearchChanged)
        self.searchFiles.clicked.connect(self.handleFileDialog)
        self.backButton.clicked.connect(self.handleBackButton)
        self.forwardButton.clicked.connect(self.handleForwardButton)
        self.bookmarkButton.clicked.connect(self.handleBookmarkButton)
        self.bookmarkMenu.triggered.connect(self.handleShowBookmark)
        self.bookmarkMenu.aboutToShow.connect(self.handleBookmarkMenu)
        self.filterButtons.buttonClicked.connect(self.updateSearchConditions)
        self.typeButtons.buttonClicked.connect(self.updateSearchConditions)
        self.otherButtons.mapped[QWidget].connect(self.updateSearchConditions)
        self.packages.expanded.connect(self.handleGroupExpanded)
        self.packages.header().sortIndicatorChanged.connect(
            self.handleSortChanged)
        self.packages.selectionModel().selectionChanged.connect(
            self.handlePackageChanged)
        self.information.currentChanged.connect(self.handleInformationChanged)
        self.details.customContextMenuRequested.connect(
            self.handleContextMenu)
        self.details.anchorClicked.connect(self.handleLinkActivated)
        self.tree.customContextMenuRequested.connect(self.handleContextMenu)
        self.tree.anchorClicked.connect(self.handleLinkActivated)
        self.files.customContextMenuRequested.connect(self.handleContextMenu)
        self.log.customContextMenuRequested.connect(self.handleContextMenu)
        self.cache.customContextMenuRequested.connect(self.handleContextMenu)
        self.backup.customContextMenuRequested.connect(self.handleContextMenu)
        self.showFindBar.triggered.connect(self.handleShowFind)
        self.findEdit.textChanged.connect(self.handleFindChanged)
        self.findNext.clicked.connect(lambda: self.handleFindNext())
        self.findPrevious.clicked.connect(self.handleFindPrevious)
        self.findClose.clicked.connect(self.findBar.hide)
        settings = qApp.settings()
        settings.beginGroup('window')
        geometry = settings.value('geometry', None, 'QByteArray')
        if geometry is not None:
            self.restoreGeometry(geometry)
        state = settings.value('central-splitter', None, 'QByteArray')
        if state is not None:
            self.centralSplitter.restoreState(state)
        state = settings.value('left-splitter', None, 'QByteArray')
        if state is not None:
            self.leftSplitter.restoreState(state)
        state = settings.value('right-splitter', None, 'QByteArray')
        if state is not None:
            self.rightSplitter.restoreState(state)
        settings.endGroup()
        settings.beginGroup('layout')
        for column, flag in enumerate(settings.value('columns', '')):
            self.packages.setColumnHidden(column, column and flag == '0')
        settings.endGroup()
        settings.beginGroup('options')
        if settings.value('work-offline', False):
            self.fileOffline.setChecked(True)
            self.handleWorkOffline(True)
        elif settings.value('include-aur', False):
            self.scopeAUR.setChecked(True)
            self.updateSearchConditions(self.scopeAUR)
        settings.endGroup()
        settings.beginGroup('search')
        strings = settings.value('strings', None, 'QStringList')
        if strings is not None:
            self.searchBox.model().setStringList(
                strings[:self.searchBox.maxCount()])
        settings.endGroup()
        settings.beginGroup('bookmarks')
        names = settings.value('names', None, 'QStringList')
        if names is not None:
            self._bookmarks.extend(names)
        settings.deleteLater()
        sys.excepthook = self.handleExceptions

    def setup(self):
        self.filters.clear()
        self.packages.model().setRowCount(0)
        self.setCurrentPackage(None)
        self.setDisabled(True, True)
        Cache.set_path(qApp.cacheDirectory())
        if self.keyFiles.isChecked():
            self.updateSearchConditions(self.keyFiles)
        self.statusBar().showMessage(
            self.tr('Initializing package data. Please wait...'))
        self.setActive(True, Callback.BackendInitialize)
        backend.initialize()
        self.updateFilters()
        self.activateSearchBox(True)

    def eventFilter(self, widget, event):
        if event.type() == QEvent.StatusTip and not len(event.tip()):
            return True
        elif (event.type() == QEvent.Shortcut and
              isinstance(widget, QGroupBox)):
            target = None
            for child in widget.children():
                if hasattr(child, 'setFocus'):
                    if target is None:
                        target = child
                    if isinstance(child, QRadioButton) and child.isChecked():
                        target = child
                        break
            if target is not None:
                target.setFocus(Qt.ShortcutFocusReason)
                return True
        elif (event.type() == QEvent.Show and
            widget is self.searchBox.view() and
            not self.searchBox.currentText()):
            view = self.searchBox.view()
            view.setCurrentIndex(view.model().index(0, 0))
        elif (self._active and isinstance(event, QInputEvent) and
             (widget is self.filters or widget is self.filters.viewport())):
            return True
        elif (event.type() == QEvent.MouseButtonPress and
              widget is self.information):
            if event.button() == Qt.XButton1:
                self.handleBackButton()
            elif event.button() == Qt.XButton2:
                self.handleForwardButton()
        elif isinstance(event, Callback):
            items, exception, duration = event.data
            del event.data
            self.setActive(False)
            self.statusBar().clearMessage()
            if event.type() == Callback.BackendInitialize:
                self.setDisabled(False, True)
            elif event.type() == Callback.ListItems:
                self.searchButton.setIcon(self.iconSearch)
                if items is not None:
                    self.filters.clearSelection()
                    self.listItems(items)
                    self.showCount(duration)
            if exception is not None:
                raise exception
            return True
        elif (widget is self.bookmarkMenu and
              event.type() == QEvent.KeyPress and
              event.key() == Qt.Key_Delete):
            action = widget.activeAction()
            if action is not None:
                widget.hide()
                if self.messageBox(self.tr('Delete Bookmark'), self.tr(
                    'Okay to delete bookmark <b>%s</b>?'
                    % action.text()), 'Yes No:', 'Question'):
                    try:
                        self._bookmarks.remove(action.text())
                    except IndexError:
                        pass
        return QMainWindow.eventFilter(self, widget, event)

    def closeEvent(self, event):
        settings = qApp.settings()
        settings.beginGroup('window')
        settings.setValue('geometry', self.saveGeometry())
        settings.setValue(
            'central-splitter', self.centralSplitter.saveState())
        settings.setValue('left-splitter', self.leftSplitter.saveState())
        settings.setValue('right-splitter', self.rightSplitter.saveState())
        settings.endGroup()
        settings.beginGroup('layout')
        settings.setValue(
            'columns', ''.join(str(item[2]) for item in self.columnInfo()))
        settings.endGroup()
        settings.beginGroup('options')
        settings.setValue('work-offline', self.fileOffline.isChecked())
        settings.endGroup()
        settings.beginGroup('search')
        settings.setValue('strings', self.searchBox.model().stringList())
        settings.endGroup()
        settings.beginGroup('bookmarks')
        settings.setValue('names', self._bookmarks)
        settings.deleteLater()
        try:
            backend.release()
        except DatabaseError:
            pass
        event.accept()

    def changeEvent(self, event):
        if event.type() == QEvent.PaletteChange:
            qApp.setOverrideCursor(Qt.BusyCursor)
            self.setCurrentPackage(self.currentPackage())
            self.updateInformation()
            qApp.restoreOverrideCursor()
        QMainWindow.changeEvent(self, event)

    def messageBox(self, title, text, buttons='Ok', icon='Information',
                   details=None, fragment=True):
        dialog = QMessageBox(self)
        dialog.setIcon(getattr(QMessageBox, icon))
        dialog.setWindowTitle('%s - %s' % (title, qApp.applicationTitle()))
        dialog.setTextFormat(Qt.RichText)
        if fragment:
            text = '<body style="white-space: pre">%s</body>' % text
        dialog.setText(text)
        if details:
            dialog.setDetailedText(details)
        for name in buttons.split():
            name, escape, default = (
                name.strip(':'), name.startswith(':'), name.endswith(':'))
            button = getattr(QMessageBox, name)
            dialog.addButton(button)
            if default:
                dialog.setDefaultButton(button)
            if escape:
                dialog.setEscapeButton(button)
        result = None
        if not dialog.exec_() & int(QMessageBox.Cancel | QMessageBox.Close):
            role = dialog.buttonRole(dialog.clickedButton())
            if role == QMessageBox.AcceptRole or role == QMessageBox.YesRole:
                result = True
            elif role == QMessageBox.RejectRole or role == QMessageBox.NoRole:
                result = False
        dialog.deleteLater()
        return result

    def openLink(self, url):
        import webbrowser
        url = QUrl(url)
        try:
            if not url.isValid():
                raise webbrowser.Error(
                    self.tr('The specified URL is invalid.'))
            if not webbrowser.open(url.toString()):
                raise webbrowser.Error(
                    self.tr('Failed to start external browser.'))
        except webbrowser.Error as exception:
            self.messageBox(
                self.tr('Open Url'), self.tr(
                '<br>Could not open url: <br><b>%s</b><br><br>%s<br>'
                % (url.toString(), exception)), 'Ok', 'Warning')

    def processException(self, cls, exception, traceback):
        if isinstance(exception, Traceback):
            print(exception)
            details = str(exception)
        else:
            sys.__excepthook__(cls, exception, traceback)
            details = Traceback.format(cls, exception, traceback)
        return details

    def handleExceptions(self, cls=None, exception=None, traceback=None):
        if not self._aborting:
            while qApp.overrideCursor() is not None:
                qApp.restoreOverrideCursor()
            if isinstance(exception, NetworkError):
                url = QUrl.fromEncoded(
                    bytes(exception.url, 'ascii')).toString()
                if len(url) > 80:
                    url = '%s ...' % url[:80]
                self.messageBox(
                    self.tr('Network Error'), self.tr(
                    '<br>Could not fetch url:<br><b>%s</b><br><br>%s<br>'
                    % (url, exception)), 'Ok', 'Warning')
            elif isinstance(exception, PatternError):
                self.messageBox(self.tr('Search'), self.tr(
                    '<br>Invalid search pattern.<br>'
                    '<br>reason: %s'
                    '<br>source: %s<br>'
                    % (exception, exception.text)), 'Ok', 'Warning')
            elif isinstance(exception, DatabaseError):
                self._aborting = int(self.messageBox(
                    self.tr('Critical Error'), self.tr(
                    '<br>Could not initialise backend:<br><br>%s.<br>'
                    % exception), 'Abort', 'Critical') is False)
            elif not isinstance(exception, KeyboardInterrupt):
                details = self.processException(cls, exception, traceback)
                url = qApp.applicationUrl('ticket')
                self._aborting = int(self.messageBox(
                    self.tr('Error'), self.tr(
                    '<br>An unhandled exception occurred.<br>'
                    '<br>Please make a <a href="%s">bug report</a> using '
                    'the details below.<br>'
                    % url), 'Ok Abort', 'Warning', details) is False)
        if self._aborting:
            if self._aborting == 1:
                self._aborting += 1
                self.messageBox(self.tr('Information'), self.tr(
                    '<br>The application will now close down.<br>'
                    ), 'Ok')
                self.close()
            else:
                try:
                    self.processException(cls, exception, traceback)
                except Exception:
                    try:
                        sys.__excepthook__(cls, exception, traceback)
                        sys.__excepthook__(*sys.exc_info())
                    except Exception:
                        print('ERROR: could not process exceptions')
                finally:
                    qApp.quit()

    def handleWorkOffline(self, on):
        backend.set_offline(on)
        if self.scopeAUR.isChecked():
            self.scopeAUR.setChecked(False)
            self.updateSearchConditions(self.scopeAUR)
        self.scopeAUR.setDisabled(
            on or self.keyFiles.isChecked() or
            self.scopeGroups.isChecked())

    def handleRefresh(self):
        self.setup()

    @delayed
    def handleStatistics(self):
        self.messageBox(self.tr('Statistics'),
                        self.format.statistics(backend.statistics()),
                        'Ok', 'NoIcon', None, False)

    def summarize(self):
        root = self.packages.model().invisibleRootItem()
        for index in range(root.rowCount()):
            yield root.child(index, 0).data(Qt.UserRole)

    def handleCopyList(self):
        text = self.format.list(self.summarize())
        qApp.clipboard().setText(text)

    @delayed
    def handleManual(self):
        if self._help is None:
            self._help = HelpDialog()
        self._help.show()

    @delayed
    def handleAbout(self):
        if self._about is None:
            self._about = AboutDialog()
        self._about.exec_()

    @delayed
    def handleAboutQt(self):
        QMessageBox.aboutQt(self, qApp.applicationTitle())

    def handleFilterActivated(self):
        selection = self.filters.selectedItems()
        if selection and not self._active:
            self.statusBar().clearMessage()
            state, location = selection[0].data(0, Qt.UserRole)
            if state & State.Group:
                self.listItems(backend.list_groups(location))
            else:
                self.listItems(backend.list_packages(state, location))

    def handleFilterExpanded(self, item=None):
        if item is None or item.childCount():
            return
        filters, location = item.data(0, Qt.UserRole)
        parent = self.addFilter(self.tr('Installed'),
            filters | State.Installed, location, item)
        self.addFilter(self.tr('Explicit'),
            filters | State.Explicit, location, parent)
        self.addFilter(self.tr('Dependency'),
            filters | State.Dependency, location, parent)
        self.addFilter(self.tr('Optional'),
            filters | State.Optional, location, parent)
        self.addFilter(self.tr('Orphan'),
            filters | State.Orphan, location, parent)
        parent.setExpanded(not location)
        self.addFilter(self.tr('Non-Installed'),
            filters | State.NonInstalled, location, item)
        self.addFilter(self.tr('Updates'),
            filters | State.Update, location, item)
        if not filters & State.Foreign:
            self.addFilter(self.tr('Groups'),
                filters | State.Group, location, item)

    def addFilter(self, title, filters=0, location=None, parent=None):
        if parent is None:
            parent = self.filters
        item = QTreeWidgetItem(parent, [title])
        item.setData(0, Qt.UserRole, (filters, location))
        if parent is self.filters and location is not None:
            item.setIcon(0, self.iconRepository)
            item.setChildIndicatorPolicy(QTreeWidgetItem.ShowIndicator)
        else:
            item.setIcon(0, self.iconFilter)
        return item

    def updateFilters(self):
        self.filters.clear()
        item = self.addFilter(self.tr('All'), location='')
        item.setExpanded(True)
        repositories = backend.list_repositories()
        local = repositories.pop()
        for name in repositories:
            self.addFilter(name.title(), location=name)
        self.addFilter(self.tr('Foreign'), State.Foreign, local)

    def updateSearchKeys(self, button, qualifier=False):
        on = button.isChecked()
        for child in self.keyGroup.findChildren(QCheckBox):
            if child is not button:
                child.setEnabled(not on and not qualifier)
                child.setChecked(False)

    def updateSearchConditions(self, button):
        if button is self.keyFiles:
            on = button.isChecked()
            self.updateSearchKeys(button)
            self.keyNames.setChecked(not on)
            self.keyDescriptions.setChecked(not on)
            self.scopePackages.setChecked(True)
            self.scopeGroups.setEnabled(not on)
            self.scopeAUR.setEnabled(
                not on and not self.fileOffline.isChecked())
            self.scopeAUR.setChecked(False)
            on = on and not Cache.has_files()
            self.scopeAll.setEnabled(not on)
            if on:
                self.scopeInstalled.setChecked(True)
            self.scopeNonInstalled.setEnabled(not on)
        elif self.typeButtons.id(button) >= 0:
            if self.scopeGroups.isEnabled():
                on = self.scopeGroups.isChecked()
                self.updateSearchKeys(self.scopeGroups)
                self.keyNames.setChecked(True)
                self.keyNames.setEnabled(True)
                self.keyDescriptions.setChecked(not on)
                self.scopeAUR.setEnabled(
                    not on and not self.fileOffline.isChecked())
                self.scopeAUR.setChecked(False)
        elif button is self.scopeAUR:
            on = button.isChecked()
            maintainer = self.keyMaintainer.isChecked()
            self.updateSearchKeys(button, maintainer)
            self.keyNames.setChecked(not maintainer)
            self.keyNames.setEnabled(not maintainer)
            self.keyDescriptions.setChecked(not maintainer)
            self.keyDescriptions.setEnabled(not maintainer)
            self.keyMaintainer.setEnabled(True)
            self.keyMaintainer.setChecked(maintainer)
            self.scopePackages.setChecked(on)
            self.scopeGroups.setEnabled(not on and not maintainer)
        elif button is self.keyMaintainer:
            on = button.isChecked()
            aur = self.scopeAUR.isChecked()
            self.updateSearchKeys(button, aur)
            self.keyNames.setChecked(not on)
            self.keyNames.setEnabled(not on)
            self.keyDescriptions.setChecked(not on)
            self.keyDescriptions.setEnabled(not on)
            self.scopePackages.setChecked(on)
            self.scopeGroups.setEnabled(not on)
        self.searchBox.setEnabled(any(
            child.isChecked() for child in self.keyGroup.children()
            if hasattr(child, 'isChecked')
            ))
        self.searchFiles.setVisible(self.keyFiles.isChecked())
        self.handleSearchChanged()

    def activateSearchBox(self, clear=False):
        if clear:
            self.searchBox.clearEditText()
        self.searchBox.setFocus()
        self.searchBox.lineEdit().selectAll()

    def handleSearchChanged(self, text=None):
        if text is None:
            text = self.searchBox.currentText()
        self.searchButton.setEnabled(
            bool(self.searchBox.isEnabled() and text))
        qApp.processEvents()

    def handleSearchActivated(self):
        if self._active:
            self.searchButton.setIcon(self.iconSearch)
            self.handleCancelTask()
            return
        text = self.searchBox.currentText()
        if not text:
            return
        self.statusBar().clearMessage()
        index = self.searchBox.findText(text)
        if index > 0:
            self.searchBox.removeItem(index)
        if index != 0:
            self.searchBox.insertItem(0, text)
            count = self.searchBox.count()
            if count > self.searchBox.maxCount():
                self.searchBox.removeItem(count - 1)
        self.searchBox.setCurrentIndex(0)
        filters = State.Installed | State.NonInstalled | State.Update
        if self.scopeInstalled.isChecked():
            filters = State.Installed
        elif self.scopeNonInstalled.isChecked():
            filters = State.NonInstalled | State.Update
        if self.scopeAUR.isChecked():
            filters |= State.AUR
        if self.scopeGroups.isChecked():
            filters |= State.Group
        keys = []
        if self.keyNames.isChecked():
            keys.append('name')
        if self.keyDescriptions.isChecked():
            keys.append('description')
        if self.keyDepends.isChecked():
            keys.append('depends')
        if self.keyProvides.isChecked():
            keys.append('provides')
        if self.keyReplaces.isChecked():
            keys.append('replaces')
        if self.keyOptional.isChecked():
            keys.append('optdepends')
        if self.keyMaintainer.isChecked():
            keys.append('maintainer')
        if self.keyFiles.isChecked():
            keys.append('files')
            filters &= ~State.AUR
        self.searchButton.setIcon(self.iconStop)
        self.setActive(True, Callback.ListItems)
        backend.find(text, filters, keys)

    def handleCancelTask(self):
        self.setActive(False)
        self.statusBar().showMessage(
            self.tr('The current task was cancelled'), 5000)

    def setDisabled(self, disabled, setup=False):
        self.fileOffline.setDisabled(disabled)
        self.fileRefresh.setDisabled(disabled)
        if setup:
            self.toolsStatistics.setDisabled(disabled)
            self.toolsCopy.setDisabled(disabled)
            self.bookmarkMenu.setDisabled(disabled)
            if disabled:
                try:
                    self.searchBox.lineEdit().returnPressed.disconnect(
                        self.handleSearchActivated)
                except TypeError:
                    pass
                try:
                    self.searchButton.clicked.disconnect(
                        self.handleSearchActivated)
                except TypeError:
                    pass
            else:
                self.searchBox.lineEdit().returnPressed.connect(
                    self.handleSearchActivated)
                self.searchButton.clicked.connect(
                    self.handleSearchActivated)

    def setActive(self, activated, identifier=None):
        self._active = activated
        self.setDisabled(activated)
        self.progress.reset()
        self.progress.setValue(0)
        if activated:
            self.progress.setRange(0, 0)
            self.progress.show()
        else:
            self.progress.setRange(0, 1)
            self.progress.hide()
        self.setCallback(identifier)

    def setCallback(self, identifier):
        if identifier is not None:
            start = time.time()
            def callback(items, exception):
                duration = time.time() - start
                qApp.postEvent(self.centralWidget(),
                    Callback(identifier, items, exception, duration))
            backend.set_callback(callback)
        else:
            backend.set_callback(None)

    def handleFileDialog(self):
        path = self.getPath('choose')
        if path is not None:
            self.searchBox.setEditText(path)

    def getPath(self, mode, filename=''):
        dialog = QFileDialog(self)
        dialog.setFilter(dialog.filter() | QDir.Hidden)
        if mode == 'choose':
            dialog.setWindowTitle(dialog.tr('Choose File'))
            dialog.setFileMode(QFileDialog.ExistingFile)
            dialog.setDirectory(QDir.root())
        elif mode == 'saveas':
            dialog.setAcceptMode(QFileDialog.AcceptSave)
            dialog.setFileMode(QFileDialog.AnyFile)
            dialog.setDirectory(QDir.home())
            dialog.selectFile(filename)
        result = None
        if dialog.exec_() == QFileDialog.Accepted:
            files = dialog.selectedFiles()
            if len(files):
                result = files[0]
        dialog.deleteLater()
        return result

    def handleGroupExpanded(self, index):
        item = self.packages.model().itemFromIndex(index)
        if item is not None and item.columnCount() <= 1:
            self.statusBar().clearMessage()
            summary = item.data(Qt.UserRole)
            packages = backend.list_group(summary.repository, summary.name)
            self.listItems(packages, item)

    def handleSortChanged(self, index=0, order=None):
        model = self.packages.model()
        item = model.horizontalHeaderItem(index)
        if order is None:
            order = item.data(Qt.InitialSortOrderRole)
        model.setSortRole(item.data())
        self.packages.sortByColumn(index, order)

    def handleInformationChanged(self, index):
        tab = self.information.widget(index)
        if tab is not None:
            tab.layout().addWidget(self.findBar)
        self.updateInformation(index)

    def handlePackageChanged(self, selected, deselected):
        row = selected.indexes()
        if row:
            item = self.packages.model().itemFromIndex(row[0])
            summary = item.data(Qt.UserRole)
            if not summary.state & State.Group:
                qApp.setOverrideCursor(Qt.BusyCursor)
                self.clearHistory()
                package = backend.get_package(summary)
                self.setCurrentPackage(package)
                self.updateInformation()
                qApp.restoreOverrideCursor()

    def handleExportInformation(self):
        widget = self.informationWidget()
        package = self.currentPackage()
        if widget is not None and package is not None:
            source = widget.objectName()
            name = '%s-%s.%s' % (package['name'], package['version'], source)
            path = self.getPath('saveas', name)
            if path is not None:
                file = QFile(path)
                if file.open(QFile.WriteOnly):
                    stream = QTextStream(file)
                    stream << self.format.file(widget.toPlainText(), source)
                    file.close()
                if file.error() != QFile.NoError:
                    self.messageBox(
                        self.tr('IO Error'), self.tr(
                        '<br>Could not save file:<br><b>%s</b><br><br>%s<br>'
                        % (path, file.errorString())), 'Ok', 'Warning')

    def handleContextMenu(self, pos=None):
        widget = self.informationWidget()
        if widget is not None:
            menu = widget.createStandardContextMenu(pos)
            menu.addSeparator()
            menu.addAction(self.showFindBar)
            menu.addSeparator()
            action = menu.addAction(self.tr('Export...'),
                self.handleExportInformation)
            action.setEnabled(not widget.document().isEmpty())
            action = menu.addAction(self.tr('Send To List'))
            targets = self.findTargets(widget, pos)
            if targets:
                action.triggered.connect(
                    lambda: self.handleSendToList(targets))
            else:
                action.setEnabled(False)
            menu.exec_(widget.viewport().mapToGlobal(pos))
            menu.deleteLater()

    def findTargets(self, browser, pos):
        targets = []
        if pos is not None:
            def find_targets(block):
                iterator = block.begin()
                while not iterator.atEnd():
                    fragment = iterator.fragment()
                    if fragment is not None:
                        format = fragment.charFormat()
                        if format.isAnchor():
                            target = format.anchorHref()
                            if target and len(format.anchorNames()):
                                yield target
                    iterator += 1
            page = browser.objectName()
            cursor = browser.cursorForPosition(pos)
            if page == 'tree':
                position = cursor.position()
                cursor.movePosition(QTextCursor.Start)
                while not cursor.atEnd():
                    table = cursor.currentTable()
                    if table is None:
                        cursor.movePosition(QTextCursor.NextBlock)
                    else:
                        cursor = table.lastCursorPosition()
                        cursor.movePosition(QTextCursor.NextBlock)
                        if position > cursor.position():
                            block = cursor.block()
                            while block.isValid():
                                targets.extend(find_targets(block))
                                block = block.next()
                        break
            elif page == 'details':
                table = cursor.currentTable()
                if table and table.cellAt(cursor).column() == 1:
                    targets.extend(find_targets(cursor.block()))
        return targets

    def handleSendToList(self, targets):
        self.setActive(True, Callback.ListItems)
        backend.list_targets(targets)

    def handleShowFind(self):
        self.findBar.show()
        self.findEdit.setFocus()

    def handleFindChanged(self, text):
        self.findNext.setEnabled(bool(text))
        self.findPrevious.setEnabled(bool(text))
        self.handleFindNext(text, True)

    def handleFindNext(self, text=None, incremental=False, backwards=False):
        widget = self.informationWidget()
        if widget is not None:
            if text is None:
                text = self.findEdit.text()
            position = 0
            found = False
            cursor = widget.textCursor()
            if text:
                flags = QTextDocument.FindFlags(0)
                if backwards:
                    flags |= QTextDocument.FindBackward
                if self.findCase.isChecked():
                    flags |= QTextDocument.FindCaseSensitively
                position = cursor.selectionStart()
                if incremental:
                    cursor.setPosition(position)
                    widget.setTextCursor(cursor)
                found = widget.find(text, flags)
                if not found:
                    if backwards:
                        cursor.movePosition(QTextCursor.End)
                    else:
                        cursor.movePosition(QTextCursor.Start)
                    widget.setTextCursor(cursor)
                    found = widget.find(text, flags)
            if not found:
                cursor.setPosition(position)
                widget.setTextCursor(cursor)
            if not found and text:
                self.findEdit.setStyleSheet(
                    'color: white; background-color: lightcoral')
            else:
                self.findEdit.setStyleSheet('')

    def handleFindPrevious(self):
        self.handleFindNext(backwards=True)

    def handleLinkActivated(self, url):
        self.navigate(url, False)

    def navigate(self, url, rebase):
        widget = self.informationWidget()
        if widget is not None:
            widget.clearFocus()
            widget.setFocus()
        if url.host():
            self.openLink(url)
        else:
            if rebase:
                self.clearHistory()
            self.setCurrentPackage(backend.get_package(url.toString()))
            self.packages.clearSelection()
            self.updateInformation()

    def clearHistory(self):
        del self._history[:]
        self._index = 0

    def updateHistory(self, name):
        if name:
            if not self._history or name != self._history[self._index]:
                del self._history[self._index + 1:]
                self._history.append(name)
                self._index = len(self._history) - 1
        else:
            self.clearHistory()
        self.backButton.setDisabled(self._index < 1)
        self.forwardButton.setDisabled(
            self._index >= len(self._history) - 1)

    def navigateHistory(self, back):
        self._index += (1 - 2 * bool(back))
        self.navigate(QUrl(self._history[self._index]), False)

    def handleBackButton(self):
        if self._index > 0:
            self.navigateHistory(True)

    def handleForwardButton(self):
        if self._index < len(self._history) - 1:
            self.navigateHistory(False)

    def handleBookmarkButton(self):
        package = self.currentPackage()
        if package is not None:
            name = package['name']
            try:
                index = self._bookmarks.index(name)
            except ValueError:
                index = -1
            else:
                if index:
                    del self._bookmarks[index]
            if index:
                self._bookmarks.insert(0, name)
                if len(self._bookmarks) > 16:
                    del self._bookmarks[-1]

    def handleShowBookmark(self, action=None):
        if action is not None:
            self.navigate(QUrl(action.text()), True)

    def handleBookmarkMenu(self):
        self.bookmarkMenu.clear()
        for bookmark in sorted(self._bookmarks):
            self.bookmarkMenu.addAction(bookmark)

    def handleClearBookmarks(self):
        self.bookmarkMenu.clear()
        del self._bookmarks[:]

    def currentPackage(self):
        return self._package

    def setCurrentPackage(self, package):
        self._package = package
        for index in range(self.information.count()):
            widget = self.information.widget(index).findChild(QTextBrowser)
            widget.clear()
            self.information.setTabEnabled(index,
                package is not None or widget.objectName() == 'details')
        self.updateHistory(package and package['name'])
        self.bookmarkButton.setEnabled(bool(package or self._bookmarks))

    def informationWidget(self, index=None):
        if index is None:
            index = self.information.currentIndex()
        tab = self.information.widget(index)
        if tab is not None and self.information.isTabEnabled(index):
            return tab.findChild(QTextBrowser)

    def updateInformation(self, index=None):
        widget = self.informationWidget(index)
        package = self.currentPackage()
        if (package is not None and widget is not None and
            widget.document().isEmpty()):
            qApp.setOverrideCursor(Qt.BusyCursor)
            widget.setHtml(
                self.format.information(package, widget.objectName()))
            qApp.restoreOverrideCursor()

    def iconFromState(self, state):
        if state & State.Orphan:
            return self.iconOrphan
        if state & State.Dependency:
            return self.iconDependency
        if state & State.Optional:
            return self.iconOptional
        if state & (State.Explicit | State.Installed):
            return self.iconInstalled
        if state & State.Upgrade:
            return self.iconUpgrade
        if state & State.Downgrade:
            return self.iconDowngrade
        return QIcon()

    def listItems(self, items, parent=None):
        if parent is None:
            parent = self.packages.model().invisibleRootItem()
        parent.setRowCount(0)
        disabled = self.packages.palette().color(
            QPalette.Disabled, QPalette.WindowText)
        for item in items:
            status = self.iconFromState(item.state)
            row = (
                QStandardItem(item.name),
                QStandardItem(item.version),
                QStandardItem(item.repository),
                QStandardItem(status, self.format.status(item.state, False)),
                QStandardItem(self.format.date(item.date, False)),
                QStandardItem(self.format.size(item.size)),
                QStandardItem(self.format.number(item.votes)),
                QStandardItem(self.format.number(item.popularity)),
                )
            row[0].setData(item, Qt.UserRole)
            row[4].setData(item.date, Qt.UserRole)
            row[5].setData(item.size, Qt.UserRole)
            row[6].setData(item.votes, Qt.UserRole)
            row[7].setData(item.popularity, Qt.UserRole)
            if item.state & State.NonInstalled:
                row[5].setForeground(disabled)
            if item.state & State.Group:
                row[0].appendRow(QStandardItem())
                row[0].setIcon(self.iconGroup)
            else:
                row[0].setIcon(self.iconPackage)
            parent.appendRow(row)
        if not parent.index().isValid():
            self.packages.scrollToTop()
        self.handleSortChanged()
        self.showCount()

    def showCount(self, duration=None):
        count = self.packages.model().invisibleRootItem().rowCount()
        if count == 0:
            message = self.tr('Found no matching items')
        elif count == 1:
            message = self.tr('Found 1 matching item')
        else:
            message = self.tr('Found %d matching items' % count)
        if duration is not None:
            message = self.tr('%s (%.3g seconds)' % (message, duration))
        self.statusBar().showMessage(message, 15000)

    def columnInfo(self):
        model = self.packages.model()
        for column in range(model.columnCount()):
            yield (column, model.horizontalHeaderItem(column).text(),
                   int(not self.packages.isColumnHidden(column)))

    def handleHeaderMenu(self, pos):
        menu = QMenu()
        for column, text, shown in self.columnInfo():
            action = menu.addAction(text)
            action.setCheckable(True)
            action.setChecked(shown)
            action.setEnabled(bool(column))
            action.setData(column)
        action = menu.exec_(self.packages.mapToGlobal(pos))
        if action is not None:
            self.packages.setColumnHidden(
                action.data(), not action.isChecked())


class Formatter(Format, QObject):
    def __init__(self, parent=None):
        Format.__init__(self)
        QObject.__init__(self, parent)

    def colours(self):
        return {
            'body-text': 'palette(text)',
            'body-back': 'palette(base)',
            'table-border': 'palette(mid)',
            'label-text': 'palette(window-text)',
            'label-back': 'palette(window)',
            'tree-lines': qApp.palette().color(
                QPalette.Disabled, QPalette.Text).name(),
            }

    def symbols(self):
        metrics = QFontMetrics(QFont('monospace'))
        symbols = {'angle': '\u251C', 'tee': '\u2570',
                   'line': '\u2500', 'stem': '\u2502'}
        for symbol in symbols.values():
            if not metrics.inFontUcs4(ord(symbol)):
                break
        else:
            return symbols
        return Format.symbols(self)

    def status(self, state, strict=True):
        if strict or state & (State.Installed | State.Update):
            return Format.status(self, state)
        return ''

    def icon(self, state):
        if state & State.Orphan:
            return ':/icons/orphan.png'
        if state & State.Dependency:
            return ':/icons/dependency.png'
        if state & State.Optional:
            return ':/icons/optional.png'
        if state & (State.Explicit | State.Installed):
            return ':/icons/installed.png'
        if state & State.Upgrade:
            return ':/icons/upgrade.png'
        if state & State.Downgrade:
            return ':/icons/downgrade.png'
        return ''


class Callback(QEvent):
    ListItems = QEvent.registerEventType()
    BackendInitialize = QEvent.registerEventType()

    def __init__(self, *args):
        QEvent.__init__(self, args[0])
        self.data = args[1:]


class AboutDialog(QDialog, Ui_AboutDialog):
    def __init__(self):
        QDialog.__init__(self, qApp.window())
        self.setupUi(self)
        self.setWindowFlags(
            self.windowFlags() & ~Qt.WindowContextHelpButtonHint)
        from sys import version as py_version
        from PyQt5.QtCore import qVersion, PYQT_VERSION_STR
        url = qApp.applicationUrl()
        title = qApp.applicationTitle()
        self.setWindowTitle('About - %s' % title)
        self.title.setText(
            """<h2>%s %s</h2>""" % (title, qApp.applicationVersion())
            )
        self.details.setText(self.tr("""
            <div align="center">
            <p>A utility for browsing pacman databases and the AUR</p>
            <p>Copyright &copy; 2010-2022, kachelaqa<br>
            &lt;kachelaqa@gmail.com&gt;</p>
            <p><a href="%s">%s</a></p>
            <p>Using:<br>Qt %s<br>Python %s<br>PyQt %s<br>libalpm %s</p>
            </div>
            """
            % (url, url, qVersion(), py_version.split()[0],
               PYQT_VERSION_STR, backend.version())
            ))
        self.information.currentChanged.connect(self.handleInformationChanged)
        self.details.linkActivated.connect(qApp.window().openLink)

    def handleInformationChanged(self, index):
        tab = self.information.widget(index)
        if tab is self.licenseTab:
            if self.license.document().isEmpty():
                stream = QFile(':/LICENSE')
                if stream.open(QFile.ReadOnly):
                    self.license.setPlainText(str(stream.readAll(), 'utf-8'))
                    stream.close()


class HelpDialog(QDialog, Ui_HelpDialog):
    def __init__(self):
        QDialog.__init__(self, qApp.window())
        self.setupUi(self)
        self.setWindowFlags(
            self.windowFlags() & ~Qt.WindowContextHelpButtonHint)
        self.setWindowTitle('Manual - %s' % qApp.applicationTitle())
        stream = QFile(':/doc/manual.html')
        if stream.open(QFile.ReadOnly):
            self.browser.setHtml(str(stream.readAll(), 'utf-8'))
            stream.close()
        self.browser.anchorClicked.connect(qApp.window().openLink)
